import pytest
from async_generator import async_generator, yield_

from geopy.exc import ConfigurationError, GeocoderQueryError
from geopy.geocoders import IGNFrance
from test.geocoders.util import BaseTestGeocoder, env
from test.proxy_server import ProxyServerThread


class TestUnitIGNFrance:

    def test_user_agent_custom(self):
        geocoder = IGNFrance(
            api_key='DUMMYKEY1234',
            username='MUSTERMANN',
            password='tops3cr3t',
            user_agent='my_user_agent/1.0'
        )
        assert geocoder.headers['User-Agent'] == 'my_user_agent/1.0'

    def test_invalid_auth_1(self):
        with pytest.raises(ConfigurationError):
            IGNFrance(api_key="a")

    def test_invalid_auth_2(self):
        with pytest.raises(ConfigurationError):
            IGNFrance(api_key="a", username="b", referer="c")

    def test_invalid_auth_3(self):
        with pytest.raises(ConfigurationError):
            IGNFrance(api_key="a", username="b")


class BaseTestIGNFrance(BaseTestGeocoder):

    async def test_invalid_query_type(self):
        with pytest.raises(GeocoderQueryError):
            self.geocoder.geocode("44109000EX0114", query_type="invalid")

    async def test_invalid_query_parcel(self):
        with pytest.raises(GeocoderQueryError):
            self.geocoder.geocode(
                "incorrect length string",
                query_type="CadastralParcel",
            )

    async def test_geocode(self):
        await self.geocode_run(
            {"query": "44109000EX0114",
             "query_type": "CadastralParcel"},
            {"latitude": 47.222482, "longitude": -1.556303},
        )

    async def test_geocode_no_result(self):
        await self.geocode_run(
            {"query": 'asdfasdfasdf'},
            {},
            expect_failure=True,
        )

    async def test_reverse_no_result(self):
        await self.reverse_run(
            # North Atlantic Ocean
            {"query": (35.173809, -37.485351)},
            {},
            expect_failure=True
        )

    async def test_geocode_with_address(self):
        await self.geocode_run(
            {"query": "Camp des Landes, 41200 VILLEFRANCHE-SUR-CHER",
             "query_type": "StreetAddress"},
            {"latitude": 47.293048,
             "longitude": 1.718985,
             "address": "le camp des landes, 41200 Villefranche-sur-Cher"},
        )

    async def test_geocode_freeform(self):
        await self.geocode_run(
            {"query": "8 rue Général Buat, Nantes",
             "query_type": "StreetAddress",
             "is_freeform": True},
            {"address": "8 r general buat , 44000 Nantes"},
        )

    async def test_geocode_position_of_interest(self):
        res = await self.geocode_run(
            {"query": "Chambéry",
             "query_type": "PositionOfInterest",
             "exactly_one": False},
            {},
        )

        addresses = [location.address for location in res]
        assert "02000 Chambry" in addresses
        assert "16420 Saint-Christophe" in addresses

    async def test_geocode_filter_by_attribute(self):
        res = await self.geocode_run(
            {"query": "Les Molettes",
             "query_type": "PositionOfInterest",
             "maximum_responses": 10,
             "filtering": '<Place type="Departement">38</Place>',
             "exactly_one": False},
            {},
        )

        departements = [location.raw['departement'] for location in res]
        unique = list(set(departements))

        assert len(unique) == 1
        assert unique[0] == "38"

    async def test_geocode_filter_by_envelope(self):
        lat_min, lng_min, lat_max, lng_max = 45.00, 5, 46, 6.40

        spatial_filtering_envelope = """
        <gml:envelope>
            <gml:pos>{lat_min} {lng_min}</gml:pos>
            <gml:pos>{lat_max} {lng_max}</gml:pos>
        </gml:envelope>
        """.format(
            lat_min=lat_min,
            lng_min=lng_min,
            lat_max=lat_max,
            lng_max=lng_max
        )

        res_spatial_filter = await self.geocode_run(
            {"query": 'Les Molettes',
             "query_type": 'PositionOfInterest',
             "maximum_responses": 10,
             "filtering": spatial_filtering_envelope,
             "exactly_one": False},
            {},
        )

        departements_spatial = list(
            {i.raw['departement'] for i in res_spatial_filter}
        )

        res_no_spatial_filter = await self.geocode_run(
            {"query": 'Les Molettes',
             "query_type": 'PositionOfInterest',
             "maximum_responses": 10,
             "exactly_one": False},
            {},
        )

        departements_no_spatial = list(
            set([
                i.raw['departement']
                for i in res_no_spatial_filter
            ])
        )

        assert len(departements_no_spatial) > len(departements_spatial)

    async def test_reverse(self):
        res = await self.reverse_run(
            {"query": '47.229554,-1.541519'},
            {},
        )
        assert res.address == '7 av camille guerin, 44000 Nantes'

    async def test_reverse_invalid_preference(self):
        with pytest.raises(GeocoderQueryError):
            self.geocoder.reverse(
                query='47.229554,-1.541519',
                reverse_geocode_preference=['a']  # invalid
            )

    async def test_reverse_preference(self):
        res = await self.reverse_run(
            {"query": '47.229554,-1.541519',
             "exactly_one": False,
             "reverse_geocode_preference": ['StreetAddress', 'PositionOfInterest']},
            {},
        )
        addresses = [location.address for location in res]
        assert "3 av camille guerin, 44000 Nantes" in addresses
        assert "5 av camille guerin, 44000 Nantes" in addresses

    async def test_reverse_by_radius(self):
        spatial_filtering_radius = """
        <gml:CircleByCenterPoint>
            <gml:pos>{coord}</gml:pos>
            <gml:radius>{radius}</gml:radius>
        </gml:CircleByCenterPoint>
        """.format(coord='48.8033333 2.3241667', radius='50')

        res_call_radius = await self.reverse_run(
            {"query": '48.8033333,2.3241667',
             "exactly_one": False,
             "maximum_responses": 10,
             "filtering": spatial_filtering_radius},
            {},
        )

        res_call = await self.reverse_run(
            {"query": '48.8033333,2.3241667',
             "exactly_one": False,
             "maximum_responses": 10},
            {},
        )

        coordinates_couples_radius = set([
            (str(location.latitude) + ' ' + str(location.longitude))
            for location in res_call_radius
        ])
        coordinates_couples = set([
            (str(location.latitude) + ' ' + str(location.longitude))
            for location in res_call
        ])

        assert coordinates_couples_radius.issubset(coordinates_couples)


@pytest.mark.skipif(
    not bool(env.get('IGNFRANCE_KEY') and env.get('IGNFRANCE_REFERER')),
    reason="No IGNFRANCE_KEY or IGNFRANCE_REFERER env variable set"
)
class TestIGNFranceApiKeyAuth(BaseTestIGNFrance):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return IGNFrance(
            api_key=env['IGNFRANCE_KEY'],
            referer=env['IGNFRANCE_REFERER'],
            timeout=10
        )


@pytest.mark.skipif(
    not bool(env.get('IGNFRANCE_USERNAME_KEY') and env.get('IGNFRANCE_USERNAME')
             and env.get('IGNFRANCE_PASSWORD')),
    reason="No IGNFRANCE_USERNAME_KEY or IGNFRANCE_USERNAME "
    "or IGNFRANCE_PASSWORD env variable set"
)
class TestIGNFranceUsernameAuth(BaseTestIGNFrance):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return IGNFrance(
            api_key=env['IGNFRANCE_USERNAME_KEY'],
            username=env['IGNFRANCE_USERNAME'],
            password=env['IGNFRANCE_PASSWORD'],
            timeout=10,
            **kwargs
        )


@pytest.mark.skipif(
    not bool(env.get('IGNFRANCE_USERNAME_KEY') and env.get('IGNFRANCE_USERNAME')
             and env.get('IGNFRANCE_PASSWORD')),
    reason="No IGNFRANCE_USERNAME_KEY or IGNFRANCE_USERNAME "
    "or IGNFRANCE_PASSWORD env variable set"
)
class TestIGNFranceUsernameAuthProxy(BaseTestGeocoder):
    proxy_timeout = 5

    @classmethod
    def make_geocoder(cls, **kwargs):
        return IGNFrance(
            api_key=env['IGNFRANCE_USERNAME_KEY'],
            username=env['IGNFRANCE_USERNAME'],
            password=env['IGNFRANCE_PASSWORD'],
            timeout=10,
            **kwargs
        )

    @pytest.fixture(scope='class', autouse=True)
    @async_generator
    async def start_proxy(_, request, class_geocoder):
        cls = request.cls
        cls.proxy_server = ProxyServerThread(timeout=cls.proxy_timeout)
        cls.proxy_server.start()
        cls.proxy_url = cls.proxy_server.get_proxy_url()
        async with cls.inject_geocoder(cls.make_geocoder(proxies=cls.proxy_url)):
            await yield_()
        cls.proxy_server.stop()
        cls.proxy_server.join()

    async def test_proxy_is_respected(self):
        assert 0 == len(self.proxy_server.requests)
        await self.geocode_run(
            {"query": "Camp des Landes, 41200 VILLEFRANCHE-SUR-CHER",
             "query_type": "StreetAddress"},
            {"latitude": 47.293048,
             "longitude": 1.718985,
             "address": "le camp des landes, 41200 Villefranche-sur-Cher"},
        )
        assert 1 == len(self.proxy_server.requests)



import pytest

from geopy.geocoders import TomTom
from test.geocoders.util import BaseTestGeocoder, env


class BaseTestTomTom(BaseTestGeocoder):

    async def test_user_agent_custom(self):
        geocoder = self.make_geocoder(
            user_agent='my_user_agent/1.0'
        )
        assert geocoder.headers['User-Agent'] == 'my_user_agent/1.0'

    async def test_geocode(self):
        location = await self.geocode_run(
            {'query': 'москва'},
            {'latitude': 55.75587, 'longitude': 37.61768},
        )
        assert 'Москва' in location.address

    async def test_reverse(self):
        location = await self.reverse_run(
            {'query': '51.5285057, -0.1369635', 'language': 'en-US'},
            {'latitude': 51.5285057, 'longitude': -0.1369635,
             "delta": 0.3},
        )
        assert 'London' in location.address
        # Russian Moscow address can be reported differently, so
        # we're querying something more ordinary, like London.
        #
        # For example, AzureMaps might return
        # `Красная площадь, 109012 Moskva` instead of the expected
        # `Красная площадь, 109012 Москва`, even when language is
        # specified explicitly as `ru-RU`. And TomTom always returns
        # the cyrillic variant, even when the `en-US` language is
        # requested.

    async def test_geocode_empty(self):
        await self.geocode_run(
            {'query': 'sldkfhdskjfhsdkhgflaskjgf'},
            {},
            expect_failure=True,
        )


@pytest.mark.skipif(
    not bool(env.get('TOMTOM_KEY')),
    reason="No TOMTOM_KEY env variable set"
)
class TestTomTom(BaseTestTomTom):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return TomTom(env['TOMTOM_KEY'], timeout=3,
                      **kwargs)



import pytest

from geopy.geocoders import MapTiler
from geopy.point import Point
from test.geocoders.util import BaseTestGeocoder, env


@pytest.mark.skipif(
    not bool(env.get('MAPTILER_KEY')),
    reason="No MAPTILER_KEY env variable set"
)
class TestMapTiler(BaseTestGeocoder):
    @classmethod
    def make_geocoder(cls, **kwargs):
        return MapTiler(api_key=env['MAPTILER_KEY'], timeout=3, **kwargs)

    async def test_geocode(self):
        await self.geocode_run(
            {"query": "435 north michigan ave, chicago il 60611 usa"},
            {"latitude": 41.890, "longitude": -87.624},
        )

    async def test_unicode_name(self):
        await self.geocode_run(
            {"query": "Stadelhoferstrasse 8, 8001 Z\u00fcrich"},
            {"latitude": 47.36649, "longitude": 8.54855},
        )

    async def test_reverse(self):
        new_york_point = Point(40.75376406311989, -73.98489005863667)
        location = await self.reverse_run(
            {"query": new_york_point},
            {"latitude": 40.7537640, "longitude": -73.98489, "delta": 1},
        )
        assert "New York" in location.address

    async def test_zero_results(self):
        await self.geocode_run(
            {"query": 'asdfasdfasdf'},
            {},
            expect_failure=True,
        )

    async def test_geocode_outside_bbox(self):
        await self.geocode_run(
            {
                "query": "435 north michigan ave, chicago il 60611 usa",
                "bbox": [[34.172684, -118.604794],
                         [34.236144, -118.500938]]
            },
            {},
            expect_failure=True,
        )

    async def test_geocode_bbox(self):
        await self.geocode_run(
            {
                "query": "435 north michigan ave, chicago il 60611 usa",
                "bbox": [Point(35.227672, -103.271484),
                         Point(48.603858, -74.399414)]
            },
            {"latitude": 41.890, "longitude": -87.624},
        )

    async def test_geocode_proximity(self):
        await self.geocode_run(
            {"query": "200 queen street", "proximity": Point(45.3, -66.1)},
            {"latitude": 44.038901, "longitude": -64.73052, "delta": 0.1},
        )

    async def test_reverse_language(self):
        zurich_point = Point(47.3723, 8.5422)
        location = await self.reverse_run(
            {"query": zurich_point, "language": "ja"},
            {"latitude": 47.3723, "longitude": 8.5422, "delta": 1},
        )
        assert "\u30c1\u30e5\u30fc\u30ea\u30c3\u30d2" in location.address

    async def test_geocode_language(self):
        location = await self.geocode_run(
            {"query": "Z\u00fcrich", "language": "ja",
             "proximity": Point(47.3723, 8.5422)},
            {"latitude": 47.3723, "longitude": 8.5422, "delta": 1},
        )
        assert "\u30c1\u30e5\u30fc\u30ea\u30c3\u30d2" in location.address

    async def test_geocode_raw(self):
        result = await self.geocode_run({"query": "New York"}, {})
        delta = 1.0
        expected = pytest.approx((-73.8784155, 40.6930727), abs=delta)
        assert expected == result.raw['center']
        assert "relation175905" == result.raw['properties']['osm_id']

    async def test_geocode_exactly_one_false(self):
        list_result = await self.geocode_run(
            {"query": "maple street", "exactly_one": False},
            {},
        )
        assert len(list_result) >= 3



"""
Geopy can calculate geodesic distance between two points using the
`geodesic distance
<https://en.wikipedia.org/wiki/Geodesics_on_an_ellipsoid>`_ or the
`great-circle distance
<https://en.wikipedia.org/wiki/Great-circle_distance>`_,
with a default of the geodesic distance available as the function
``geopy.distance.distance``.

Great-circle distance (:class:`.great_circle`) uses a spherical model of
the earth, using the mean earth radius as defined by the International
Union of Geodesy and Geophysics, (2\\ *a* + *b*)/3 = 6371.0087714150598
kilometers approx 6371.009 km (for WGS-84), resulting in an error of up
to about 0.5%. The radius value is stored in
:const:`distance.EARTH_RADIUS`, so it can be customized (it should
always be in kilometers, however).

The geodesic distance is the shortest distance on the surface of an
ellipsoidal model of the earth.  The default algorithm uses the method
is given by `Karney (2013)
<https://doi.org/10.1007%2Fs00190-012-0578-z>`_ (:class:`.geodesic`);
this is accurate to round-off and always converges.

``geopy.distance.distance`` currently uses :class:`.geodesic`.

There are multiple popular ellipsoidal models,
and which one will be the most accurate depends on where your points are
located on the earth.  The default is the WGS-84 ellipsoid, which is the
most globally accurate.  geopy includes a few other models in the
:const:`distance.ELLIPSOIDS` dictionary::

                  model             major (km)   minor (km)     flattening
    ELLIPSOIDS = {'WGS-84':        (6378.137,    6356.7523142,  1 / 298.257223563),
                  'GRS-80':        (6378.137,    6356.7523141,  1 / 298.257222101),
                  'Airy (1830)':   (6377.563396, 6356.256909,   1 / 299.3249646),
                  'Intl 1924':     (6378.388,    6356.911946,   1 / 297.0),
                  'Clarke (1880)': (6378.249145, 6356.51486955, 1 / 293.465),
                  'GRS-67':        (6378.1600,   6356.774719,   1 / 298.25),
                  }

Here are examples of ``distance.distance`` usage::

    >>> from geopy import distance
    >>> newport_ri = (41.49008, -71.312796)
    >>> cleveland_oh = (41.499498, -81.695391)
    >>> print(distance.distance(newport_ri, cleveland_oh).miles)
    538.39044536

    >>> wellington = (-41.32, 174.81)
    >>> salamanca = (40.96, -5.50)
    >>> print(distance.distance(wellington, salamanca).km)
    19959.6792674

Using :class:`.great_circle` distance::

    >>> print(distance.great_circle(newport_ri, cleveland_oh).miles)
    536.997990696

You can change the ellipsoid model used by the geodesic formulas like so::

    >>> ne, cl = newport_ri, cleveland_oh
    >>> print(distance.geodesic(ne, cl, ellipsoid='GRS-80').miles)

The above model name will automatically be retrieved from the
:const:`distance.ELLIPSOIDS` dictionary. Alternatively, you can specify
the model values directly::

    >>> distance.geodesic(ne, cl, ellipsoid=(6377., 6356., 1 / 297.)).miles

Distances support simple arithmetic, making it easy to do things like
calculate the length of a path::

    >>> from geopy import Nominatim
    >>> d = distance.distance
    >>> g = Nominatim(user_agent="specify_your_app_name_here")
    >>> _, wa = g.geocode('Washington, DC')
    >>> _, pa = g.geocode('Palo Alto, CA')
    >>> print((d(ne, cl) + d(cl, wa) + d(wa, pa)).miles)
    3277.30439191


.. _distance_altitudes:

Currently all algorithms assume that altitudes of the points are either
zero (as in the examples above) or equal, and are relatively small.
Thus altitudes never affect the resulting distances::

    >>> from geopy import distance
    >>> newport_ri = (41.49008, -71.312796)
    >>> cleveland_oh = (41.499498, -81.695391)
    >>> print(distance.distance(newport_ri, cleveland_oh).km)
    866.4554329098687
    >>> newport_ri = (41.49008, -71.312796, 100)
    >>> cleveland_oh = (41.499498, -81.695391, 100)
    >>> print(distance.distance(newport_ri, cleveland_oh).km)
    866.4554329098687

If you need to calculate distances with elevation, then for short
distances the `Euclidean distance
<https://en.wikipedia.org/wiki/Euclidean_distance>`_ formula might give
a suitable approximation::

    >>> import math
    >>> from geopy import distance
    >>> p1 = (43.668613, 40.258916, 0.976)
    >>> p2 = (43.658852, 40.250839, 1.475)
    >>> flat_distance = distance.distance(p1[:2], p2[:2]).km
    >>> print(flat_distance)
    1.265133525952866
    >>> euclidian_distance = math.sqrt(flat_distance**2 + (p2[2] - p1[2])**2)
    >>> print(euclidian_distance)
    1.359986705262199

An attempt to calculate distances between points with different altitudes
would result in a ``ValueError`` exception.

"""
import abc
from math import asin, atan2, cos, sin, sqrt

from geographiclib.geodesic import Geodesic

from geopy import units, util
from geopy.point import Point
from geopy.units import radians

# IUGG mean earth radius in kilometers, from
# https://en.wikipedia.org/wiki/Earth_radius#Mean_radius.  Using a
# sphere with this radius results in an error of up to about 0.5%.
EARTH_RADIUS = 6371.009

# From http://www.movable-type.co.uk/scripts/LatLongVincenty.html:
#   The most accurate and widely used globally-applicable model for the earth
#   ellipsoid is WGS-84, used in this script. Other ellipsoids offering a
#   better fit to the local geoid include Airy (1830) in the UK, International
#   1924 in much of Europe, Clarke (1880) in Africa, and GRS-67 in South
#   America. America (NAD83) and Australia (GDA) use GRS-80, functionally
#   equivalent to the WGS-84 ellipsoid.
ELLIPSOIDS = {
    # model           major (km)   minor (km)     flattening
    'WGS-84':        (6378.137, 6356.7523142, 1 / 298.257223563),
    'GRS-80':        (6378.137, 6356.7523141, 1 / 298.257222101),
    'Airy (1830)':   (6377.563396, 6356.256909, 1 / 299.3249646),
    'Intl 1924':     (6378.388, 6356.911946, 1 / 297.0),
    'Clarke (1880)': (6378.249145, 6356.51486955, 1 / 293.465),
    'GRS-67':        (6378.1600, 6356.774719, 1 / 298.25)
}


def cmp(a, b):
    return (a > b) - (a < b)


def lonlat(x, y, z=0):
    """
    ``geopy.distance.distance`` accepts coordinates in ``(y, x)``/``(lat, lon)``
    order, while some other libraries and systems might use
    ``(x, y)``/``(lon, lat)``.

    This function provides a convenient way to convert coordinates of the
    ``(x, y)``/``(lon, lat)`` format to a :class:`geopy.point.Point` instance.

    Example::

        >>> from geopy.distance import lonlat, distance
        >>> newport_ri_xy = (-71.312796, 41.49008)
        >>> cleveland_oh_xy = (-81.695391, 41.499498)
        >>> print(distance(lonlat(*newport_ri_xy), lonlat(*cleveland_oh_xy)).miles)
        538.3904453677203

    :param x: longitude
    :param y: latitude
    :param z: (optional) altitude
    :return: Point(latitude, longitude, altitude)
    """
    return Point(y, x, z)


def _ensure_same_altitude(a, b):
    if abs(a.altitude - b.altitude) > 1e-6:
        raise ValueError(
            'Calculating distance between points with different altitudes '
            'is not supported'
        )
    # Note: non-zero equal altitudes are fine: assuming that
    # the elevation is many times smaller than the Earth radius
    # it won't give much error.


class Distance(abc.ABC):

    def __init__(self, *args, **kwargs):
        kilometers = kwargs.pop('kilometers', 0)
        if len(args) == 1:
            # if we only get one argument we assume
            # it's a known distance instead of
            # calculating it first
            kilometers += args[0]
        elif len(args) > 1:
            for a, b in util.pairwise(args):
                kilometers += self.measure(a, b)

        kilometers += units.kilometers(**kwargs)
        self.__kilometers = kilometers

    def __add__(self, other):
        if isinstance(other, Distance):
            return self.__class__(self.kilometers + other.kilometers)
        else:
            raise TypeError(
                "Distance instance must be added with Distance instance."
            )

    def __neg__(self):
        return self.__class__(-self.kilometers)

    def __sub__(self, other):
        return self + -other

    def __mul__(self, other):
        return self.__class__(self.kilometers * other)

    def __div__(self, other):
        if isinstance(other, Distance):
            return self.kilometers / other.kilometers
        else:
            return self.__class__(self.kilometers / other)

    __truediv__ = __div__

    def __abs__(self):
        return self.__class__(abs(self.kilometers))

    def __nonzero__(self):
        return bool(self.kilometers)

    __bool__ = __nonzero__

    @abc.abstractmethod
    def measure(self, a, b):
        raise NotImplementedError()

    def __repr__(self):  # pragma: no cover
        return 'Distance(%s)' % self.kilometers

    def __str__(self):  # pragma: no cover
        return '%s km' % self.__kilometers

    def __cmp__(self, other):  # py2 only
        if isinstance(other, Distance):
            return cmp(self.kilometers, other.kilometers)
        else:
            return cmp(self.kilometers, other)

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __gt__(self, other):
        return self.__cmp__(other) > 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __ge__(self, other):
        return self.__cmp__(other) >= 0

    def __le__(self, other):
        return self.__cmp__(other) <= 0

    @property
    def kilometers(self):
        return self.__kilometers

    @property
    def km(self):
        return self.kilometers

    @property
    def meters(self):
        return units.meters(kilometers=self.kilometers)

    @property
    def m(self):
        return self.meters

    @property
    def miles(self):
        return units.miles(kilometers=self.kilometers)

    @property
    def mi(self):
        return self.miles

    @property
    def feet(self):
        return units.feet(kilometers=self.kilometers)

    @property
    def ft(self):
        return self.feet

    @property
    def nautical(self):
        return units.nautical(kilometers=self.kilometers)

    @property
    def nm(self):
        return self.nautical


class great_circle(Distance):
    """
    Use spherical geometry to calculate the surface distance between two
    points.

    Set which radius of the earth to use by specifying a ``radius`` keyword
    argument. It must be in kilometers. The default is to use the module
    constant `EARTH_RADIUS`, which uses the average great-circle radius.

    Example::

        >>> from geopy.distance import great_circle
        >>> newport_ri = (41.49008, -71.312796)
        >>> cleveland_oh = (41.499498, -81.695391)
        >>> print(great_circle(newport_ri, cleveland_oh).miles)
        536.997990696

    """

    def __init__(self, *args, **kwargs):
        self.RADIUS = kwargs.pop('radius', EARTH_RADIUS)
        super().__init__(*args, **kwargs)

    def measure(self, a, b):
        a, b = Point(a), Point(b)
        _ensure_same_altitude(a, b)

        lat1, lng1 = radians(degrees=a.latitude), radians(degrees=a.longitude)
        lat2, lng2 = radians(degrees=b.latitude), radians(degrees=b.longitude)

        sin_lat1, cos_lat1 = sin(lat1), cos(lat1)
        sin_lat2, cos_lat2 = sin(lat2), cos(lat2)

        delta_lng = lng2 - lng1
        cos_delta_lng, sin_delta_lng = cos(delta_lng), sin(delta_lng)

        d = atan2(sqrt((cos_lat2 * sin_delta_lng) ** 2 +
                       (cos_lat1 * sin_lat2 -
                        sin_lat1 * cos_lat2 * cos_delta_lng) ** 2),
                  sin_lat1 * sin_lat2 + cos_lat1 * cos_lat2 * cos_delta_lng)

        return self.RADIUS * d

    def destination(self, point, bearing, distance=None):
        """
        TODO docs.
        """
        point = Point(point)
        lat1 = units.radians(degrees=point.latitude)
        lng1 = units.radians(degrees=point.longitude)
        bearing = units.radians(degrees=bearing)

        if distance is None:
            distance = self
        if isinstance(distance, Distance):
            distance = distance.kilometers

        d_div_r = float(distance) / self.RADIUS

        lat2 = asin(
            sin(lat1) * cos(d_div_r) +
            cos(lat1) * sin(d_div_r) * cos(bearing)
        )

        lng2 = lng1 + atan2(
            sin(bearing) * sin(d_div_r) * cos(lat1),
            cos(d_div_r) - sin(lat1) * sin(lat2)
        )

        return Point(units.degrees(radians=lat2), units.degrees(radians=lng2))


GreatCircleDistance = great_circle


class geodesic(Distance):
    """
    Calculate the geodesic distance between two points.

    Set which ellipsoidal model of the earth to use by specifying an
    ``ellipsoid`` keyword argument. The default is 'WGS-84', which is the
    most globally accurate model.  If ``ellipsoid`` is a string, it is
    looked up in the `ELLIPSOIDS` dictionary to obtain the major and minor
    semiaxes and the flattening. Otherwise, it should be a tuple with those
    values.  See the comments above the `ELLIPSOIDS` dictionary for
    more information.

    Example::

        >>> from geopy.distance import geodesic
        >>> newport_ri = (41.49008, -71.312796)
        >>> cleveland_oh = (41.499498, -81.695391)
        >>> print(geodesic(newport_ri, cleveland_oh).miles)
        538.390445368

    """

    ellipsoid_key = None
    ELLIPSOID = None
    geod = None

    def __init__(self, *args, **kwargs):
        self.set_ellipsoid(kwargs.pop('ellipsoid', 'WGS-84'))
        major, minor, f = self.ELLIPSOID
        super().__init__(*args, **kwargs)

    def set_ellipsoid(self, ellipsoid):
        """
        Change the ellipsoid used in the calculation.
        """
        if isinstance(ellipsoid, str):
            try:
                self.ELLIPSOID = ELLIPSOIDS[ellipsoid]
                self.ellipsoid_key = ellipsoid
            except KeyError:
                raise Exception(
                    "Invalid ellipsoid. See geopy.distance.ELLIPSOIDS"
                )
        else:
            self.ELLIPSOID = ellipsoid
            self.ellipsoid_key = None

    # Call geographiclib routines for measure and destination
    def measure(self, a, b):
        a, b = Point(a), Point(b)
        _ensure_same_altitude(a, b)
        lat1, lon1 = a.latitude, a.longitude
        lat2, lon2 = b.latitude, b.longitude

        if not (isinstance(self.geod, Geodesic) and
                self.geod.a == self.ELLIPSOID[0] and
                self.geod.f == self.ELLIPSOID[2]):
            self.geod = Geodesic(self.ELLIPSOID[0], self.ELLIPSOID[2])

        s12 = self.geod.Inverse(lat1, lon1, lat2, lon2,
                                Geodesic.DISTANCE)['s12']

        return s12

    def destination(self, point, bearing, distance=None):
        """
        TODO docs.
        """
        point = Point(point)
        lat1 = point.latitude
        lon1 = point.longitude
        azi1 = bearing

        if distance is None:
            distance = self
        if isinstance(distance, Distance):
            distance = distance.kilometers

        if not (isinstance(self.geod, Geodesic) and
                self.geod.a == self.ELLIPSOID[0] and
                self.geod.f == self.ELLIPSOID[2]):
            self.geod = Geodesic(self.ELLIPSOID[0], self.ELLIPSOID[2])

        r = self.geod.Direct(lat1, lon1, azi1, distance,
                             Geodesic.LATITUDE | Geodesic.LONGITUDE)

        return Point(r['lat2'], r['lon2'])


GeodesicDistance = geodesic

# Set the default distance formula
distance = GeodesicDistance



import warnings

import pytest

from geopy import exc
from geopy.geocoders import Here
from geopy.point import Point
from test.geocoders.util import BaseTestGeocoder, env


class TestUnitHere:

    def test_user_agent_custom(self):
        geocoder = Here(
            apikey='DUMMYKEY1234',
            user_agent='my_user_agent/1.0'
        )
        assert geocoder.headers['User-Agent'] == 'my_user_agent/1.0'

    def test_error_with_no_keys(self):
        with pytest.raises(exc.ConfigurationError):
            Here()

    def test_warning_with_legacy_auth(self):
        with warnings.catch_warnings(record=True) as w:
            Here(
                app_id='DUMMYID1234',
                app_code='DUMMYCODE1234',
            )
        assert len(w) == 1

    def test_no_warning_with_apikey(self):
        with warnings.catch_warnings(record=True) as w:
            Here(
                apikey='DUMMYKEY1234',
            )
        assert len(w) == 0


class BaseTestHere(BaseTestGeocoder):

    async def test_geocode_empty_result(self):
        await self.geocode_run(
            {"query": "xqj37"},
            {},
            expect_failure=True
        )

    async def test_geocode(self):
        await self.geocode_run(
            {"query": "435 north michigan ave, chicago il 60611 usa"},
            {"latitude": 41.890, "longitude": -87.624}
        )

    async def test_geocode_structured(self):
        query = {
            "street": "north michigan ave",
            "housenumber": "435",
            "city": "chicago",
            "state": "il",
            "postalcode": 60611,
            "country": "usa"
        }
        await self.geocode_run(
            {"query": query},
            {"latitude": 41.890, "longitude": -87.624}
        )

    async def test_geocode_unicode_name(self):
        # unicode in Japanese for Paris. (POIs not included.)
        await self.geocode_run(
            {"query": "\u30d1\u30ea"},
            {"latitude": 48.85718, "longitude": 2.34141}
        )

    async def test_bbox(self):
        await self.geocode_run(
            {"query": "moscow",  # Idaho USA
             "bbox": [[50.1, -130.1], [44.1, -100.9]]},
            {"latitude": 46.7323875, "longitude": -117.0001651},
        )

    async def test_mapview(self):
        await self.geocode_run(
            {"query": "moscow",  # Idaho USA
             "mapview": [[50.1, -130.1], [44.1, -100.9]]},
            {"latitude": 46.7323875, "longitude": -117.0001651},
        )

    async def test_geocode_shapes(self):
        address_string = "435 north michigan ave, chicago il 60611 usa"
        res = await self.geocode_run(
            {"query": address_string, "additional_data": "IncludeShapeLevel,postalCode"},
            {"latitude": 41.89035, "longitude": -87.62333},
        )
        shape_value = res.raw['Location']['Shape']['Value']
        assert shape_value.startswith('MULTIPOLYGON (((')

    async def test_geocode_with_language_de(self):
        address_string = "435 north michigan ave, chicago il 60611 usa"
        res = await self.geocode_run(
            {"query": address_string, "language": "de-DE"},
            {}
        )
        assert "Vereinigte Staaten" in res.address

    async def test_geocode_with_language_en(self):
        address_string = "435 north michigan ave, chicago il 60611 usa"
        res = await self.geocode_run(
            {"query": address_string, "language": "en-US"},
            {}
        )
        assert "United States" in res.address

    async def test_geocode_with_paging(self):
        address_string = "Hauptstr., Berlin, Germany"
        input = {"query": address_string, "maxresults": 12, "exactly_one": False}
        res = await self.geocode_run(input, {})
        assert len(res) == 12

        input["pageinformation"] = 2
        res = await self.geocode_run(input, {})
        assert len(res) >= 3
        assert len(res) <= 6

        input["pageinformation"] = 3
        res = await self.geocode_run(input, {}, expect_failure=True)

    async def test_reverse(self):
        await self.reverse_run(
            {"query": Point(40.753898, -73.985071)},
            {"latitude": 40.753898, "longitude": -73.985071}
        )

    async def test_reverse_point_radius_1000_float(self):
        # needs more testing
        res = await self.reverse_run(
            {"query": Point(40.753898, -73.985071), "radius": 1000.12,
             "exactly_one": False},
            {"latitude": 40.753898, "longitude": -73.985071}
        )
        assert len(res) > 5

    async def test_reverse_point_radius_10(self):
        # needs more testing
        res = await self.reverse_run(
            {"query": Point(40.753898, -73.985071), "radius": 10, "exactly_one": False},
            {"latitude": 40.753898, "longitude": -73.985071}
        )
        assert len(res) > 5

    async def test_reverse_with_language_de(self):
        res = await self.reverse_run(
            {"query": Point(40.753898, -73.985071), "language": "de-DE"},
            {}
        )
        assert "Vereinigte Staaten" in res.address

    async def test_reverse_with_language_en(self):
        res = await self.reverse_run(
            {"query": Point(40.753898, -73.985071), "language": "en-US"},
            {}
        )
        assert "United States" in res.address

    async def test_reverse_with_mode_areas(self):
        res = await self.reverse_run(
            {"query": Point(40.753898, -73.985071), "mode": "retrieveAreas"},
            {}
        )
        assert "Theater District-Times Square" in res.address

    async def test_reverse_with_maxresults_5(self):
        res = await self.reverse_run(
            {
                "query": Point(40.753898, -73.985071),
                "maxresults": 5,
                "exactly_one": False
            },
            {}
        )
        assert len(res) == 5


@pytest.mark.skipif(
    not bool(env.get('HERE_APIKEY')),
    reason="No HERE_APIKEY env variable set"
)
class TestHereApiKey(BaseTestHere):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return Here(
            apikey=env['HERE_APIKEY'],
            timeout=10,
            **kwargs
        )


@pytest.mark.skipif(
    not bool(env.get('HERE_APP_ID') and env.get('HERE_APP_CODE')),
    reason="No HERE_APP_ID or HERE_APP_CODE env variable set"
)
class TestHereLegacyAuth(BaseTestHere):

    @classmethod
    def make_geocoder(cls, **kwargs):
        with warnings.catch_warnings(record=True) as w:
            geocoder = Here(
                app_id=env['HERE_APP_ID'],
                app_code=env['HERE_APP_CODE'],
                timeout=10,
                **kwargs
            )
        assert len(w) == 1
        return geocoder



from geopy.geocoders import AlgoliaPlaces
from geopy.point import Point
from test.geocoders.util import BaseTestGeocoder, env


class TestAlgoliaPlaces(BaseTestGeocoder):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return AlgoliaPlaces(
            app_id=env.get('ALGOLIA_PLACES_APP_ID'),
            api_key=env.get('ALGOLIA_PLACES_API_KEY'),
            timeout=3,
            **kwargs)

    async def test_user_agent_custom(self):
        geocoder = self.make_geocoder(
            user_agent='my_user_agent/1.0'
        )
        assert geocoder.headers['User-Agent'] == 'my_user_agent/1.0'

    async def test_geocode(self):
        location = await self.geocode_run(
            {'query': 'москва'},
            {'latitude': 55.75587, 'longitude': 37.61768},
        )
        assert 'Москва' in location.address

    async def test_reverse(self):
        location = await self.reverse_run(
            {'query': '51, -0.13', 'language': 'en'},
            {'latitude': 51, 'longitude': -0.13},
        )
        assert 'A272' in location.address

    async def test_reverse_no_result(self):
        await self.reverse_run(
            # North Atlantic Ocean
            {'query': (35.173809, -37.485351)},
            {},
            expect_failure=True
        )

    async def test_explicit_type(self):
        location = await self.geocode_run(
            {'query': 'Madrid', 'type': 'city', 'language': 'en'},
            {},
        )
        assert 'Madrid' in location.address

    async def test_limit(self):
        limit = 5
        locations = await self.geocode_run(
            {'query': 'Madrid', 'type': 'city',
             'language': 'en', 'exactly_one': False,
             'limit': limit},
            {},
        )
        assert len(locations) == limit

    async def test_countries(self):
        countries = ["ES"]
        location = await self.geocode_run(
            {'query': 'Madrid', 'language': 'en',
             'countries': countries},
            {},
        )
        assert "Madrid" in location.address

    async def test_countries_no_result(self):
        countries = ["NO", "IT"]
        await self.geocode_run(
            {'query': 'Madrid', 'language': 'en',
             'countries': countries},
            {},
            expect_failure=True
        )

    async def test_geocode_no_result(self):
        await self.geocode_run(
            {'query': 'sldkfhdskjfhsdkhgflaskjgf'},
            {},
            expect_failure=True,
        )

    async def test_around(self):
        await self.geocode_run(
            {'query': 'maple street', 'language': 'en', 'around': Point(51.1, -0.1)},
            {'latitude': 51.5299, 'longitude': -0.0628044, "delta": 1},
        )
        await self.geocode_run(
            {'query': 'maple street', 'language': 'en', 'around': Point(50.1, 10.1)},
            {'latitude': 50.0517, 'longitude': 10.1966, "delta": 1},
        )



from geopy.geocoders import BANFrance
from test.geocoders.util import BaseTestGeocoder


class TestBANFrance(BaseTestGeocoder):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return BANFrance(timeout=10, **kwargs)

    async def test_user_agent_custom(self):
        geocoder = BANFrance(
            user_agent='my_user_agent/1.0'
        )
        assert geocoder.headers['User-Agent'] == 'my_user_agent/1.0'

    async def test_geocode_with_address(self):
        location = await self.geocode_run(
            {"query": "Camp des Landes, 41200 VILLEFRANCHE-SUR-CHER"},
            {"latitude": 47.293048, "longitude": 1.718985},
        )
        assert "Camp des Landes" in location.address

    async def test_reverse(self):
        location = await self.reverse_run(
            {"query": "48.154587,3.221237"},
            {"latitude": 48.154587, "longitude": 3.221237},
        )
        assert "Rue des Fontaines" in location.address

    async def test_geocode_limit(self):
        result = await self.geocode_run(
            {"query": "8 bd du port", "limit": 2, "exactly_one": False},
            {}
        )
        assert 2 >= len(result)



from unittest.mock import patch

import pytest

from geopy import exc
from geopy.geocoders import GeocodeFarm
from geopy.point import Point
from test.geocoders.util import BaseTestGeocoder, env


@pytest.mark.xfail(
    env.get('GEOCODEFARM_SKIP'),
    reason=(
        "geocodefarm service is unstable at times"
    )
)
class TestGeocodeFarm(BaseTestGeocoder):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return GeocodeFarm(
            # None api_key will use free tier on GeocodeFarm.
            api_key=env.get('GEOCODEFARM_KEY'),
            timeout=10,
            **kwargs
        )

    async def test_user_agent_custom(self):
        geocoder = GeocodeFarm(
            user_agent='my_user_agent/1.0'
        )
        assert geocoder.headers['User-Agent'] == 'my_user_agent/1.0'

    async def test_geocode(self):
        location = await self.geocode_run(
            {"query": "435 north michigan ave, chicago il 60611 usa"},
            {"latitude": 41.890, "longitude": -87.624},
        )
        assert "chicago" in location.address.lower()

    async def test_location_address(self):
        await self.geocode_run(
            {"query": "moscow"},
            {"address": "Moscow, Russia",
             "latitude": 55.7558913503453, "longitude": 37.6172961632184}
        )

    async def test_reverse(self):
        location = await self.reverse_run(
            {"query": Point(40.75376406311989, -73.98489005863667)},
            {"latitude": 40.75376406311989, "longitude": -73.98489005863667},
            skiptest_on_failure=True,  # sometimes the result is empty
        )
        assert "new york" in location.address.lower()

    async def test_authentication_failure(self):
        async with self.inject_geocoder(GeocodeFarm(api_key="invalid")):
            with pytest.raises(exc.GeocoderAuthenticationFailure):
                await self.geocode_run(
                    {"query": '435 north michigan ave, chicago il 60611'},
                    {},
                    expect_failure=True,
                )

    async def test_quota_exceeded(self):

        def mock_call_geocoder(url, callback, **kwargs):
            return callback({
                "geocoding_results": {
                    "STATUS": {
                        "access": "OVER_QUERY_LIMIT",
                        "status": "FAILED, ACCESS_DENIED"
                    }
                }
            })

        with patch.object(self.geocoder, '_call_geocoder', mock_call_geocoder):
            with pytest.raises(exc.GeocoderQuotaExceeded):
                self.geocoder.geocode('435 north michigan ave, chicago il 60611')

    async def test_no_results(self):
        await self.geocode_run(
            {"query": "gibberish kdjhsakdjh skjdhsakjdh"},
            {},
            expect_failure=True
        )

    async def test_unhandled_api_error(self):

        def mock_call_geocoder(url, callback, **kwargs):
            return callback({
                "geocoding_results": {
                    "STATUS": {
                        "access": "BILL_PAST_DUE",
                        "status": "FAILED, ACCESS_DENIED"
                    }
                }
            })

        with patch.object(self.geocoder, '_call_geocoder', mock_call_geocoder):
            with pytest.raises(exc.GeocoderServiceError):
                self.geocoder.geocode('435 north michigan ave, chicago il 60611')



""":class:`.RateLimiter` and :class:`.AsyncRateLimiter` allow to perform bulk
operations while gracefully handling error responses and adding delays
when needed.

In the example below a delay of 1 second (``min_delay_seconds=1``)
will be added between each pair of ``geolocator.geocode`` calls; all
:class:`geopy.exc.GeocoderServiceError` exceptions will be retried
(up to ``max_retries`` times)::

    import pandas as pd
    df = pd.DataFrame({'name': ['paris', 'berlin', 'london']})

    from geopy.geocoders import Nominatim
    geolocator = Nominatim(user_agent="specify_your_app_name_here")

    from geopy.extra.rate_limiter import RateLimiter
    geocode = RateLimiter(geolocator.geocode, min_delay_seconds=1)
    df['location'] = df['name'].apply(geocode)

    df['point'] = df['location'].apply(lambda loc: tuple(loc.point) if loc else None)

This would produce the following DataFrame::

    >>> df
         name                                           location  \\
    0   paris  (Paris, Île-de-France, France métropolitaine, ...
    1  berlin  (Berlin, 10117, Deutschland, (52.5170365, 13.3...
    2  london  (London, Greater London, England, SW1A 2DU, UK...

                               point
    0   (48.8566101, 2.3514992, 0.0)
    1  (52.5170365, 13.3888599, 0.0)
    2  (51.5073219, -0.1276474, 0.0)

To pass extra options to the `geocode` call::

    from functools import partial
    df['location'] = df['name'].apply(partial(geocode, language='de'))

To see a progress bar::

    from tqdm import tqdm
    tqdm.pandas()
    df['location'] = df['name'].progress_apply(geocode)

Before using rate limiting classes, please consult with the Geocoding
service ToS, which might explicitly consider bulk requests (even throttled)
a violation.
"""

import asyncio
import inspect
import threading
from itertools import chain, count
from time import sleep
from timeit import default_timer

from geopy.exc import GeocoderServiceError
from geopy.util import logger

__all__ = ("AsyncRateLimiter", "RateLimiter")


def _is_last_gen(count):
    """list(_is_last_gen(2)) -> [False, False, True]"""
    return chain((False for _ in range(count)), [True])


class BaseRateLimiter:
    """Base Rate Limiter class for both sync and async versions."""

    _retry_exceptions = (GeocoderServiceError,)

    def __init__(
        self,
        *,
        min_delay_seconds,
        max_retries,
        swallow_exceptions,
        return_value_on_exception
    ):
        self.min_delay_seconds = min_delay_seconds
        self.max_retries = max_retries
        self.swallow_exceptions = swallow_exceptions
        self.return_value_on_exception = return_value_on_exception
        assert max_retries >= 0

        # State:
        self._lock = threading.Lock()
        self._last_call = None

    def _clock(self):  # pragma: no cover
        return default_timer()

    def _acquire_request_slot_gen(self):
        # Requests rate is limited by `min_delay_seconds` interval.
        #
        # Imagine the time axis as a grid with `min_delay_seconds` step,
        # where we would call each step as a "request slot". RateLimiter
        # guarantees that each "request slot" contains at most 1 request.
        #
        # Note that actual requests might take longer time than
        # `min_delay_seconds`. In that case you might want to consider
        # parallelizing requests (with a ThreadPool for sync mode and
        # asyncio tasks for async), to keep the requests rate closer
        # to `min_delay_seconds`.
        #
        # This generator thread-safely acquires a "request slot", and
        # if it fails to do that at this time, it yields the amount
        # of seconds to sleep until the next attempt. The generator
        # stops only when the "request slot" has been successfully
        # acquired.
        #
        # There's no ordering between the concurrent requests. The first
        # request to acquire the lock wins the next "request slot".
        while True:
            with self._lock:
                clock = self._clock()
                if self._last_call is None:
                    # A first iteration -- start immediately.
                    self._last_call = clock
                    return
                seconds_since_last_call = clock - self._last_call
                wait = self.min_delay_seconds - seconds_since_last_call
                if wait <= 0:
                    # A successfully acquired request slot.
                    self._last_call = clock
                    return
            # Couldn't acquire a request slot. Wait until the beginning
            # of the next slot to try again.
            yield wait

    def _retries_gen(self, args, kwargs):
        for i, is_last_try in zip(count(), _is_last_gen(self.max_retries)):
            try:
                yield i  # Run the function.
            except self._retry_exceptions:
                if is_last_try:
                    yield True  # The exception should be raised
                else:
                    logger.warning(
                        type(self).__name__ + " caught an error, retrying "
                        "(%s/%s tries). Called with (*%r, **%r).",
                        i,
                        self.max_retries,
                        args,
                        kwargs,
                        exc_info=True,
                    )
                    yield False  # The exception has been swallowed.
                    continue
            else:
                # A successful run -- stop retrying:
                return  # pragma: no cover

    def _handle_exc(self, args, kwargs):
        if self.swallow_exceptions:
            logger.warning(
                type(self).__name__ + " swallowed an error after %r retries. "
                "Called with (*%r, **%r).",
                self.max_retries,
                args,
                kwargs,
                exc_info=True,
            )
            return self.return_value_on_exception
        else:
            raise


class RateLimiter(BaseRateLimiter):
    """This is a Rate Limiter implementation for synchronous functions
    (like geocoders with the default :class:`geopy.adapters.BaseSyncAdapter`).

    Examples::

        from geopy.extra.rate_limiter import RateLimiter
        from geopy.geocoders import Nominatim

        geolocator = Nominatim(user_agent="specify_your_app_name_here")

        search = ["moscow", "paris", "berlin", "tokyo", "beijing"]
        geocode = RateLimiter(geolocator.geocode, min_delay_seconds=1)
        locations = [geocode(s) for s in search]

        search = [
            (55.47, 37.32), (48.85, 2.35), (52.51, 13.38),
            (34.69, 139.40), (39.90, 116.39)
        ]
        reverse = RateLimiter(geolocator.reverse, min_delay_seconds=1)
        locations = [reverse(s) for s in search]

    RateLimiter class is thread-safe. If geocoding service's responses
    are slower than `min_delay_seconds`, then you can benefit from
    parallelizing the work::

        import concurrent.futures

        geolocator = OpenMapQuest(api_key="...")
        geocode = RateLimiter(geolocator.geocode, min_delay_seconds=1/20)

        with concurrent.futures.ThreadPoolExecutor() as e:
            locations = list(e.map(geocode, search))

    .. versionchanged:: 2.0
       Added thread-safety support.
    """

    def __init__(
        self,
        func,
        *,
        min_delay_seconds=0.0,
        max_retries=2,
        error_wait_seconds=5.0,
        swallow_exceptions=True,
        return_value_on_exception=None
    ):
        """
        :param callable func:
            A function which should be wrapped by the rate limiter.

        :param float min_delay_seconds:
            Minimum delay in seconds between the wrapped ``func`` calls.
            To convert :abbr:`RPS (Requests Per Second)` rate to
            ``min_delay_seconds`` you need to divide 1 by RPS. For example,
            if you need to keep the rate at 20 RPS, you can use
            ``min_delay_seconds=1/20``.

        :param int max_retries:
            Number of retries on exceptions. Only
            :class:`geopy.exc.GeocoderServiceError` exceptions are
            retried -- others are always re-raised. ``max_retries + 1``
            requests would be performed at max per query. Set
            ``max_retries=0`` to disable retries.

        :param float error_wait_seconds:
            Time to wait between retries after errors. Must be
            greater or equal to ``min_delay_seconds``.

        :param bool swallow_exceptions:
            Should an exception be swallowed after retries? If not,
            it will be re-raised. If yes, the ``return_value_on_exception``
            will be returned.

        :param return_value_on_exception:
            Value to return on failure when ``swallow_exceptions=True``.

        """
        super().__init__(
            min_delay_seconds=min_delay_seconds,
            max_retries=max_retries,
            swallow_exceptions=swallow_exceptions,
            return_value_on_exception=return_value_on_exception,
        )
        self.func = func
        self.error_wait_seconds = error_wait_seconds
        assert error_wait_seconds >= min_delay_seconds
        assert max_retries >= 0

    def _sleep(self, seconds):  # pragma: no cover
        logger.debug(type(self).__name__ + " sleep(%r)", seconds)
        sleep(seconds)

    def _acquire_request_slot(self):
        for wait in self._acquire_request_slot_gen():
            self._sleep(wait)

    def __call__(self, *args, **kwargs):
        gen = self._retries_gen(args, kwargs)
        for _ in gen:
            self._acquire_request_slot()
            try:
                res = self.func(*args, **kwargs)
                if inspect.isawaitable(res):
                    raise ValueError(
                        "An async awaitable has been passed to `RateLimiter`. "
                        "Use `AsyncRateLimiter` instead, which supports awaitables."
                    )
                return res
            except self._retry_exceptions as e:
                if gen.throw(e):
                    # A final try
                    return self._handle_exc(args, kwargs)
                self._sleep(self.error_wait_seconds)

        raise RuntimeError("Should not have been reached")  # pragma: no cover


class AsyncRateLimiter(BaseRateLimiter):
    """This is a Rate Limiter implementation for asynchronous functions
    (like geocoders with :class:`geopy.adapters.BaseAsyncAdapter`).

    Examples::

        from geopy.adapters import AioHTTPAdapter
        from geopy.extra.rate_limiter import AsyncRateLimiter
        from geopy.geocoders import Nominatim

        async with Nominatim(
            user_agent="specify_your_app_name_here",
            adapter_factory=AioHTTPAdapter,
        ) as geolocator:

            search = ["moscow", "paris", "berlin", "tokyo", "beijing"]
            geocode = AsyncRateLimiter(geolocator.geocode, min_delay_seconds=1)
            locations = [await geocode(s) for s in search]

            search = [
                (55.47, 37.32), (48.85, 2.35), (52.51, 13.38),
                (34.69, 139.40), (39.90, 116.39)
            ]
            reverse = AsyncRateLimiter(geolocator.reverse, min_delay_seconds=1)
            locations = [await reverse(s) for s in search]

    AsyncRateLimiter class is safe to use across multiple concurrent tasks.
    If geocoding service's responses are slower than `min_delay_seconds`,
    then you can benefit from parallelizing the work::

        import asyncio

        async with OpenMapQuest(
            api_key="...", adapter_factory=AioHTTPAdapter
        ) as geolocator:

            geocode = AsyncRateLimiter(geolocator.geocode, min_delay_seconds=1/20)
            locations = await asyncio.gather(*(geocode(s) for s in search))

    .. versionadded:: 2.0
    """

    def __init__(
        self,
        func,
        *,
        min_delay_seconds=0.0,
        max_retries=2,
        error_wait_seconds=5.0,
        swallow_exceptions=True,
        return_value_on_exception=None
    ):
        """
        :param callable func:
            A function which should be wrapped by the rate limiter.

        :param float min_delay_seconds:
            Minimum delay in seconds between the wrapped ``func`` calls.
            To convert :abbr:`RPS (Requests Per Second)` rate to
            ``min_delay_seconds`` you need to divide 1 by RPS. For example,
            if you need to keep the rate at 20 RPS, you can use
            ``min_delay_seconds=1/20``.

        :param int max_retries:
            Number of retries on exceptions. Only
            :class:`geopy.exc.GeocoderServiceError` exceptions are
            retried -- others are always re-raised. ``max_retries + 1``
            requests would be performed at max per query. Set
            ``max_retries=0`` to disable retries.

        :param float error_wait_seconds:
            Time to wait between retries after errors. Must be
            greater or equal to ``min_delay_seconds``.

        :param bool swallow_exceptions:
            Should an exception be swallowed after retries? If not,
            it will be re-raised. If yes, the ``return_value_on_exception``
            will be returned.

        :param return_value_on_exception:
            Value to return on failure when ``swallow_exceptions=True``.

        """
        super().__init__(
            min_delay_seconds=min_delay_seconds,
            max_retries=max_retries,
            swallow_exceptions=swallow_exceptions,
            return_value_on_exception=return_value_on_exception,
        )
        self.func = func
        self.error_wait_seconds = error_wait_seconds
        assert error_wait_seconds >= min_delay_seconds
        assert max_retries >= 0

    async def _sleep(self, seconds):  # pragma: no cover
        logger.debug(type(self).__name__ + " sleep(%r)", seconds)
        await asyncio.sleep(seconds)

    async def _acquire_request_slot(self):
        for wait in self._acquire_request_slot_gen():
            await self._sleep(wait)

    async def __call__(self, *args, **kwargs):
        gen = self._retries_gen(args, kwargs)
        for _ in gen:
            await self._acquire_request_slot()
            try:
                return await self.func(*args, **kwargs)
            except self._retry_exceptions as e:
                if gen.throw(e):
                    # A final try
                    return self._handle_exc(args, kwargs)
                await self._sleep(self.error_wait_seconds)

        raise RuntimeError("Should not have been reached")  # pragma: no cover



from functools import partial
from urllib.parse import urlencode

from geopy.exc import GeocoderParseError, GeocoderServiceError
from geopy.geocoders.base import DEFAULT_SENTINEL, Geocoder
from geopy.location import Location
from geopy.util import logger

__all__ = ("Yandex", )


class Yandex(Geocoder):
    """Yandex geocoder.

    Documentation at:
        https://tech.yandex.com/maps/doc/geocoder/desc/concepts/input_params-docpage/

    .. attention::
        Since September 2019 Yandex requires each request to have an API key.
        API keys can be created at https://developer.tech.yandex.ru/
    """

    api_path = '/1.x/'

    def __init__(
            self,
            api_key,
            *,
            timeout=DEFAULT_SENTINEL,
            proxies=DEFAULT_SENTINEL,
            user_agent=None,
            scheme=None,
            ssl_context=DEFAULT_SENTINEL,
            adapter_factory=None
    ):
        """

        :param str api_key: Yandex API key, mandatory.
            The key can be created at https://developer.tech.yandex.ru/

        :param int timeout:
            See :attr:`geopy.geocoders.options.default_timeout`.

        :param dict proxies:
            See :attr:`geopy.geocoders.options.default_proxies`.

        :param str user_agent:
            See :attr:`geopy.geocoders.options.default_user_agent`.

        :param str scheme:
            See :attr:`geopy.geocoders.options.default_scheme`.

        :type ssl_context: :class:`ssl.SSLContext`
        :param ssl_context:
            See :attr:`geopy.geocoders.options.default_ssl_context`.

        :param callable adapter_factory:
            See :attr:`geopy.geocoders.options.default_adapter_factory`.

            .. versionadded:: 2.0
        """
        super().__init__(
            scheme=scheme,
            timeout=timeout,
            proxies=proxies,
            user_agent=user_agent,
            ssl_context=ssl_context,
            adapter_factory=adapter_factory,
        )
        self.api_key = api_key
        domain = 'geocode-maps.yandex.ru'
        self.api = '%s://%s%s' % (self.scheme, domain, self.api_path)

    def geocode(
            self,
            query,
            *,
            exactly_one=True,
            timeout=DEFAULT_SENTINEL,
            lang=None
    ):
        """
        Return a location point by address.

        :param str query: The address or query you wish to geocode.

        :param bool exactly_one: Return one result or a list of results, if
            available.

        :param int timeout: Time, in seconds, to wait for the geocoding service
            to respond before raising a :class:`geopy.exc.GeocoderTimedOut`
            exception. Set this only if you wish to override, on this call
            only, the value set during the geocoder's initialization.

        :param str lang: Language of the response and regional settings
            of the map. List of supported values:

            - ``tr_TR`` -- Turkish (only for maps of Turkey);
            - ``en_RU`` -- response in English, Russian map features;
            - ``en_US`` -- response in English, American map features;
            - ``ru_RU`` -- Russian (default);
            - ``uk_UA`` -- Ukrainian;
            - ``be_BY`` -- Belarusian.

        :rtype: ``None``, :class:`geopy.location.Location` or a list of them, if
            ``exactly_one=False``.
        """
        params = {
            'geocode': query,
            'format': 'json'
        }
        params['apikey'] = self.api_key
        if lang:
            params['lang'] = lang
        if exactly_one:
            params['results'] = 1
        url = "?".join((self.api, urlencode(params)))
        logger.debug("%s.geocode: %s", self.__class__.__name__, url)
        callback = partial(self._parse_json, exactly_one=exactly_one)
        return self._call_geocoder(url, callback, timeout=timeout)

    def reverse(
            self,
            query,
            *,
            exactly_one=True,
            timeout=DEFAULT_SENTINEL,
            kind=None,
            lang=None
    ):
        """
        Return an address by location point.

        :param query: The coordinates for which you wish to obtain the
            closest human-readable addresses.
        :type query: :class:`geopy.point.Point`, list or tuple of ``(latitude,
            longitude)``, or string as ``"%(latitude)s, %(longitude)s"``.

        :param bool exactly_one: Return one result or a list of results, if
            available.

        :param int timeout: Time, in seconds, to wait for the geocoding service
            to respond before raising a :class:`geopy.exc.GeocoderTimedOut`
            exception. Set this only if you wish to override, on this call
            only, the value set during the geocoder's initialization.

        :param str kind: Type of toponym. Allowed values: `house`, `street`, `metro`,
            `district`, `locality`.

        :param str lang: Language of the response and regional settings
            of the map. List of supported values:

            - ``tr_TR`` -- Turkish (only for maps of Turkey);
            - ``en_RU`` -- response in English, Russian map features;
            - ``en_US`` -- response in English, American map features;
            - ``ru_RU`` -- Russian (default);
            - ``uk_UA`` -- Ukrainian;
            - ``be_BY`` -- Belarusian.

        :rtype: ``None``, :class:`geopy.location.Location` or a list of them, if
            ``exactly_one=False``.
        """

        try:
            point = self._coerce_point_to_string(query, "%(lon)s,%(lat)s")
        except ValueError:
            raise ValueError("Must be a coordinate pair or Point")
        params = {
            'geocode': point,
            'format': 'json'
        }
        params['apikey'] = self.api_key
        if lang:
            params['lang'] = lang
        if kind:
            params['kind'] = kind
        url = "?".join((self.api, urlencode(params)))
        logger.debug("%s.reverse: %s", self.__class__.__name__, url)
        callback = partial(self._parse_json, exactly_one=exactly_one)
        return self._call_geocoder(url, callback, timeout=timeout)

    def _parse_json(self, doc, exactly_one):
        """
        Parse JSON response body.
        """
        if doc.get('error'):
            raise GeocoderServiceError(doc['error']['message'])

        try:
            places = doc['response']['GeoObjectCollection']['featureMember']
        except KeyError:
            raise GeocoderParseError('Failed to parse server response')

        def parse_code(place):
            """
            Parse each record.
            """
            try:
                place = place['GeoObject']
            except KeyError:
                raise GeocoderParseError('Failed to parse server response')

            longitude, latitude = [
                float(_) for _ in place['Point']['pos'].split(' ')
            ]

            name_elements = ['name', 'description']
            location = ', '.join([place[k] for k in name_elements if place.get(k)])

            return Location(location, (latitude, longitude), place)

        if exactly_one:
            try:
                return parse_code(places[0])
            except IndexError:
                return None
        else:
            return [parse_code(place) for place in places]



"""
Each geolocation service you might use, such as Google Maps, Bing Maps, or
Nominatim, has its own class in ``geopy.geocoders`` abstracting the service's
API. Geocoders each define at least a ``geocode`` method, for resolving a
location from a string, and may define a ``reverse`` method, which resolves a
pair of coordinates to an address. Each Geocoder accepts any credentials
or settings needed to interact with its service, e.g., an API key or
locale, during its initialization.

To geolocate a query to an address and coordinates:

    >>> from geopy.geocoders import Nominatim
    >>> geolocator = Nominatim(user_agent="specify_your_app_name_here")
    >>> location = geolocator.geocode("175 5th Avenue NYC")
    >>> print(location.address)
    Flatiron Building, 175, 5th Avenue, Flatiron, New York, NYC, New York, ...
    >>> print((location.latitude, location.longitude))
    (40.7410861, -73.9896297241625)
    >>> print(location.raw)
    {'place_id': '9167009604', 'type': 'attraction', ...}


To find the address corresponding to a set of coordinates:

    >>> from geopy.geocoders import Nominatim
    >>> geolocator = Nominatim(user_agent="specify_your_app_name_here")
    >>> location = geolocator.reverse("52.509669, 13.376294")
    >>> print(location.address)
    Potsdamer Platz, Mitte, Berlin, 10117, Deutschland, European Union
    >>> print((location.latitude, location.longitude))
    (52.5094982, 13.3765983)
    >>> print(location.raw)
    {'place_id': '654513', 'osm_type': 'node', ...}

Locators' ``geocode`` and ``reverse`` methods require the argument ``query``,
and also accept at least the argument ``exactly_one``, which is ``True`` by
default.
Geocoders may have additional attributes, e.g., Bing accepts ``user_location``,
the effect of which is to bias results near that location. ``geocode``
and ``reverse`` methods  may return three types of values:

- When there are no results found, returns ``None``.

- When the method's ``exactly_one`` argument is ``True`` and at least one
  result is found, returns a :class:`geopy.location.Location` object, which
  can be iterated over as:

    ``(address<String>, (latitude<Float>, longitude<Float>))``

  Or can be accessed as ``location.address``, ``location.latitude``,
  ``location.longitude``, ``location.altitude``, and ``location.raw``. The
  last contains the full geocoder's response for this result.

- When ``exactly_one`` is ``False``, and there is at least one result, returns a
  list of :class:`geopy.location.Location` objects, as above:

    ``[location, [...]]``

If a service is unavailable or otherwise returns a non-OK response, or doesn't
receive a response in the allotted timeout, you will receive one of the
`Exceptions`_ detailed below.

.. _specifying_parameters_once:

Specifying Parameters Once
~~~~~~~~~~~~~~~~~~~~~~~~~~

Geocoding methods accept a lot of different parameters, and you would
probably want to specify some of them just once and not care about them
later.

This is easy to achieve with Python's :func:`functools.partial`::

    >>> from functools import partial
    >>> from geopy.geocoders import Nominatim

    >>> geolocator = Nominatim(user_agent="specify_your_app_name_here")

    >>> geocode = partial(geolocator.geocode, language="es")
    >>> print(geocode("london"))
    Londres, Greater London, Inglaterra, SW1A 2DX, Gran Bretaña
    >>> print(geocode("paris"))
    París, Isla de Francia, Francia metropolitana, Francia
    >>> print(geocode("paris", language="en"))
    Paris, Ile-de-France, Metropolitan France, France

    >>> reverse = partial(geolocator.reverse, language="es")
    >>> print(reverse("52.509669, 13.376294"))
    Steinecke, Potsdamer Platz, Tiergarten, Mitte, 10785, Alemania

If you need to modify the query, you can also use a one-liner with lambda.
For example, if you only need to geocode locations in `Cleveland, Ohio`,
you could do::

    >>> geocode = lambda query: geolocator.geocode("%s, Cleveland OH" % query)
    >>> print(geocode("11111 Euclid Ave"))
    Thwing Center, Euclid Avenue, Magnolia-Wade Park Historic District,
    University Circle, Cleveland, Cuyahoga County, Ohio, 44106, United States
    of America

That lambda doesn't accept kwargs. If you need them, you could do::

    >>> _geocode = partial(geolocator.geocode, language="es")
    >>> geocode = lambda query, **kw: _geocode("%s, Cleveland OH" % query, **kw)
    >>> print(geocode("11111 Euclid Ave"))
    Thwing Center, Euclid Avenue, Magnolia-Wade Park Historic District,
    University Circle, Cleveland, Cuyahoga County, Ohio, 44106, Estados Unidos
    >>> print(geocode("11111 Euclid Ave", language="en"))
    Thwing Center, Euclid Avenue, Magnolia-Wade Park Historic District,
    University Circle, Cleveland, Cuyahoga County, Ohio, 44106, United States
    of America

Geopy Is Not a Service
~~~~~~~~~~~~~~~~~~~~~~

Geocoding is provided by a number of different services, which are not
affiliated with geopy in any way. These services provide APIs, which anyone
could implement, and geopy is just a library which provides these
implementations for many different services in a single package.

.. image:: ./_static/geopy_and_geocoding_services.svg
   :target: ./_static/geopy_and_geocoding_services.svg

Therefore:

1. Different services have different Terms of Use, quotas, pricing,
   geodatabases and so on. For example, :class:`.Nominatim`
   is free, but provides low request limits. If you need to make more queries,
   consider using another (probably paid) service, such as
   :class:`.OpenMapQuest` or :class:`.PickPoint`
   (these two are commercial providers of Nominatim, so they should
   have the same data and APIs). Or, if you are ready to wait, you can try
   :mod:`geopy.extra.rate_limiter`.

2. geopy cannot be responsible for the geocoding services' databases.
   If you have issues with some queries which the service cannot fulfill,
   it should be directed to that service's support team.

3. geopy cannot be responsible for any networking issues between your computer
   and the geocoding service.

If you face any problem with your current geocoding service provider, you can
always try a different one.

.. _async_mode:

Async Mode
~~~~~~~~~~

By default geopy geocoders are synchronous (i.e. they use an Adapter
based on :class:`.BaseSyncAdapter`).

All geocoders can be used with asyncio by simply switching to an
Adapter based on :class:`.BaseAsyncAdapter` (like :class:`.AioHTTPAdapter`).

Example::

    from geopy.adapters import AioHTTPAdapter
    from geopy.geocoders import Nominatim

    async with Nominatim(
        user_agent="specify_your_app_name_here",
        adapter_factory=AioHTTPAdapter,
    ) as geolocator:
        location = await geolocator.geocode("175 5th Avenue NYC")
        print(location.address)

Basically the usage is the same as in synchronous mode, except that
all geocoder calls should be used with ``await``, and the geocoder
instance should be created by ``async with``. The context manager is optional,
however, it is strongly advised to use it to avoid resources leaks.

"""

__all__ = (
    "get_geocoder_for_service",
    "options",
    # The order of classes below should correspond to the order of their
    # files in the ``geocoders`` directory ordered by name.
    #
    # If you're adding a new geocoder class, then you should mention it in
    # this module 3 times:
    # 1. In this ``__all__`` tuple.
    # 2. In the imports block below.
    # 3. In the ``SERVICE_TO_GEOCODER`` dict below.
    #
    # Also don't forget to pull up the list of geocoders
    # in the docs: docs/index.rst
    "AlgoliaPlaces",
    "ArcGIS",
    "AzureMaps",
    "Baidu",
    "BaiduV3",
    "BANFrance",
    "Bing",
    "DataBC",
    "GeocodeEarth",
    "GeocodeFarm",
    "GeoNames",
    "GoogleV3",
    "Geolake",
    "Here",
    "IGNFrance",
    "MapBox",
    "MapQuest",
    "MapTiler",
    "Nominatim",
    "OpenCage",
    "OpenMapQuest",
    "PickPoint",
    "Pelias",
    "Photon",
    "LiveAddress",
    "TomTom",
    "What3Words",
    "Yandex",
)


from geopy.exc import GeocoderNotFound
from geopy.geocoders.algolia import AlgoliaPlaces
from geopy.geocoders.arcgis import ArcGIS
from geopy.geocoders.azure import AzureMaps
from geopy.geocoders.baidu import Baidu, BaiduV3
from geopy.geocoders.banfrance import BANFrance
from geopy.geocoders.base import options
from geopy.geocoders.bing import Bing
from geopy.geocoders.databc import DataBC
from geopy.geocoders.geocodeearth import GeocodeEarth
from geopy.geocoders.geocodefarm import GeocodeFarm
from geopy.geocoders.geolake import Geolake
from geopy.geocoders.geonames import GeoNames
from geopy.geocoders.googlev3 import GoogleV3
from geopy.geocoders.here import Here
from geopy.geocoders.ignfrance import IGNFrance
from geopy.geocoders.mapbox import MapBox
from geopy.geocoders.mapquest import MapQuest
from geopy.geocoders.maptiler import MapTiler
from geopy.geocoders.nominatim import Nominatim
from geopy.geocoders.opencage import OpenCage
from geopy.geocoders.openmapquest import OpenMapQuest
from geopy.geocoders.pelias import Pelias
from geopy.geocoders.photon import Photon
from geopy.geocoders.pickpoint import PickPoint
from geopy.geocoders.smartystreets import LiveAddress
from geopy.geocoders.tomtom import TomTom
from geopy.geocoders.what3words import What3Words
from geopy.geocoders.yandex import Yandex

SERVICE_TO_GEOCODER = {
    "algolia": AlgoliaPlaces,
    "arcgis": ArcGIS,
    "azure": AzureMaps,
    "baidu": Baidu,
    "baiduv3": BaiduV3,
    "banfrance": BANFrance,
    "bing": Bing,
    "databc": DataBC,
    "geocodeearth": GeocodeEarth,
    "geocodefarm": GeocodeFarm,
    "geonames": GeoNames,
    "google": GoogleV3,
    "googlev3": GoogleV3,
    "geolake": Geolake,
    "here": Here,
    "ignfrance": IGNFrance,
    "mapbox": MapBox,
    "mapquest": MapQuest,
    "maptiler": MapTiler,
    "nominatim": Nominatim,
    "opencage": OpenCage,
    "openmapquest": OpenMapQuest,
    "pickpoint": PickPoint,
    "pelias": Pelias,
    "photon": Photon,
    "liveaddress": LiveAddress,
    "tomtom": TomTom,
    "what3words": What3Words,
    "yandex": Yandex,
}


def get_geocoder_for_service(service):
    """
    For the service provided, try to return a geocoder class.

    >>> from geopy.geocoders import get_geocoder_for_service
    >>> get_geocoder_for_service("nominatim")
    geopy.geocoders.nominatim.Nominatim

    If the string given is not recognized, a
    :class:`geopy.exc.GeocoderNotFound` exception is raised.

    Given that almost all of the geocoders provide the ``geocode``
    method it could be used to make basic queries based entirely
    on user input::

        from geopy.geocoders import get_geocoder_for_service

        def geocode(geocoder, config, query):
            cls = get_geocoder_for_service(geocoder)
            geolocator = cls(**config)
            location = geolocator.geocode(query)
            return location.address

        >>> geocode("nominatim", dict(user_agent="specify_your_app_name_here"), \
"london")
        'London, Greater London, England, SW1A 2DX, United Kingdom'
        >>> geocode("photon", dict(), "london")
        'London, SW1A 2DX, London, England, United Kingdom'

    """
    try:
        return SERVICE_TO_GEOCODER[service.lower()]
    except KeyError:
        raise GeocoderNotFound(
            "Unknown geocoder '%s'; options are: %s" %
            (service, SERVICE_TO_GEOCODER.keys())
        )



import pytest

from geopy.exc import GeocoderAuthenticationFailure
from geopy.geocoders import Baidu, BaiduV3
from geopy.point import Point
from test.geocoders.util import BaseTestGeocoder, env


class TestUnitBaidu(BaseTestGeocoder):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return Baidu(
            api_key='DUMMYKEY1234',
            user_agent='my_user_agent/1.0',
            **kwargs
        )

    async def test_user_agent_custom(self):
        assert self.geocoder.headers['User-Agent'] == 'my_user_agent/1.0'


class BaseTestBaidu(BaseTestGeocoder):

    async def test_basic_address(self):
        await self.geocode_run(
            {"query": (
                "\u5317\u4eac\u5e02\u6d77\u6dc0\u533a"
                "\u4e2d\u5173\u6751\u5927\u885727\u53f7"
            )},
            {"latitude": 39.983615544507, "longitude": 116.32295155093},
        )

    async def test_reverse_point(self):
        await self.reverse_run(
            {"query": Point(39.983615544507, 116.32295155093)},
            {"latitude": 39.983615544507, "longitude": 116.32295155093},
        )
        await self.reverse_run(
            {"query": Point(39.983615544507, 116.32295155093), "exactly_one": False},
            {"latitude": 39.983615544507, "longitude": 116.32295155093},
        )


@pytest.mark.skipif(
    not bool(env.get('BAIDU_KEY')),
    reason="No BAIDU_KEY env variable set"
)
class TestBaidu(BaseTestBaidu):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return Baidu(
            api_key=env['BAIDU_KEY'],
            timeout=3,
            **kwargs,
        )

    async def test_invalid_ak(self):
        async with self.inject_geocoder(Baidu(api_key='DUMMYKEY1234')):
            with pytest.raises(GeocoderAuthenticationFailure) as exc_info:
                await self.geocode_run({"query": "baidu"}, None)
            assert str(exc_info.value) == 'Invalid AK'


@pytest.mark.skipif(
    not bool(env.get('BAIDU_KEY_REQUIRES_SK') and env.get('BAIDU_SEC_KEY')),
    reason="BAIDU_KEY_REQUIRES_SK and BAIDU_SEC_KEY env variables not set"
)
class TestBaiduSK(BaseTestBaidu):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return Baidu(
            api_key=env['BAIDU_KEY_REQUIRES_SK'],
            security_key=env['BAIDU_SEC_KEY'],
            timeout=3,
            **kwargs,
        )

    async def test_sn_with_peculiar_chars(self):
        await self.geocode_run(
            {"query": (
                "\u5317\u4eac\u5e02\u6d77\u6dc0\u533a"
                "\u4e2d\u5173\u6751\u5927\u885727\u53f7"
                " ' & = , ? %"
            )},
            {"latitude": 39.983615544507, "longitude": 116.32295155093},
        )


@pytest.mark.skipif(
    not bool(env.get('BAIDU_V3_KEY')),
    reason="No BAIDU_V3_KEY env variable set"
)
class TestBaiduV3(TestBaidu):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return BaiduV3(
            api_key=env['BAIDU_V3_KEY'],
            timeout=3,
            **kwargs,
        )


@pytest.mark.skipif(
    not bool(env.get('BAIDU_V3_KEY_REQUIRES_SK') and env.get('BAIDU_V3_SEC_KEY')),
    reason="BAIDU_V3_KEY_REQUIRES_SK and BAIDU_V3_SEC_KEY env variables not set"
)
class TestBaiduV3SK(TestBaiduSK):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return BaiduV3(
            api_key=env['BAIDU_V3_KEY_REQUIRES_SK'],
            security_key=env['BAIDU_V3_SEC_KEY'],
            timeout=3,
            **kwargs,
        )



import json
import os
from abc import ABC, abstractmethod
from collections import defaultdict
from unittest.mock import ANY, patch

import pytest
from async_generator import async_generator, asynccontextmanager, yield_

from geopy import exc
from geopy.adapters import BaseAsyncAdapter
from geopy.location import Location

env = defaultdict(lambda: None)
try:
    with open(".test_keys") as fp:
        env.update(json.loads(fp.read()))
except IOError:
    env.update(os.environ)


class BaseTestGeocoder(ABC):
    """
    Base for geocoder-specific test cases.
    """

    geocoder = None
    delta = 0.5

    @pytest.fixture(scope='class', autouse=True)
    @async_generator
    async def class_geocoder(_, request, patch_adapter):
        """Prepare a class-level Geocoder instance."""
        cls = request.cls
        geocoder = cls.make_geocoder()
        cls.geocoder = geocoder

        run_async = isinstance(geocoder.adapter, BaseAsyncAdapter)
        if run_async:
            async with geocoder:
                await yield_(geocoder)
        else:
            await yield_(geocoder)

    @classmethod
    @asynccontextmanager
    @async_generator
    async def inject_geocoder(cls, geocoder):
        """An async context manager allowing to inject a custom
        geocoder instance in a single test method which will
        be used by the `geocode_run`/`reverse_run` methods.
        """
        with patch.object(cls, 'geocoder', geocoder):
            run_async = isinstance(geocoder.adapter, BaseAsyncAdapter)
            if run_async:
                async with geocoder:
                    await yield_(geocoder)
            else:
                await yield_(geocoder)

    @pytest.fixture(autouse=True)
    def ensure_no_geocoder_assignment(self):
        yield
        assert self.geocoder is type(self).geocoder, (
            "Detected `self.geocoder` assignment. "
            "Please use `async with inject_geocoder(my_geocoder):` "
            "instead, whish supports async adapters."
        )

    @classmethod
    @abstractmethod
    def make_geocoder(cls, **kwargs):  # pragma: no cover
        pass

    async def geocode_run(
        self, payload, expected, expect_failure=False,
        skiptest_on_failure=False
    ):
        """
        Calls geocoder.geocode(**payload), then checks against `expected`.
        """
        cls = type(self)
        result = await self._make_request(self.geocoder, 'geocode', **payload)
        if result is None:
            if expect_failure:
                return
            elif skiptest_on_failure:
                pytest.skip('%s: Skipping test due to empty result' % cls.__name__)
            else:
                pytest.fail('%s: No result found' % cls.__name__)
        if result == []:
            pytest.fail('%s returned an empty list instead of None' % cls.__name__)
        self._verify_request(result, exactly_one=payload.get('exactly_one', True),
                             **expected)
        return result

    async def reverse_run(
        self, payload, expected, expect_failure=False,
        skiptest_on_failure=False
    ):
        """
        Calls geocoder.reverse(**payload), then checks against `expected`.
        """
        cls = type(self)
        result = await self._make_request(self.geocoder, 'reverse', **payload)
        if result is None:
            if expect_failure:
                return
            elif skiptest_on_failure:
                pytest.skip('%s: Skipping test due to empty result' % cls.__name__)
            else:
                pytest.fail('%s: No result found' % cls.__name__)
        if result == []:
            pytest.fail('%s returned an empty list instead of None' % cls.__name__)
        self._verify_request(result, exactly_one=payload.get('exactly_one', True),
                             **expected)
        return result

    async def reverse_timezone_run(self, payload, expected):
        timezone = await self._make_request(
            self.geocoder, 'reverse_timezone', **payload)
        if expected is None:
            assert timezone is None
        else:
            assert timezone.pytz_timezone == expected

        return timezone

    async def _make_request(self, geocoder, method, *args, **kwargs):
        cls = type(self)
        call = getattr(geocoder, method)
        run_async = isinstance(geocoder.adapter, BaseAsyncAdapter)
        try:
            if run_async:
                result = await call(*args, **kwargs)
            else:
                result = call(*args, **kwargs)
        except exc.GeocoderQuotaExceeded:
            pytest.skip("%s: Quota exceeded" % cls.__name__)
        except exc.GeocoderTimedOut:
            pytest.skip("%s: Service timed out" % cls.__name__)
        except exc.GeocoderUnavailable:
            pytest.skip("%s: Service unavailable" % cls.__name__)
        return result

    def _verify_request(
            self,
            result,
            latitude=ANY,
            longitude=ANY,
            address=ANY,
            exactly_one=True,
            delta=None,
    ):
        if exactly_one:
            assert isinstance(result, Location)
        else:
            assert isinstance(result, list)

        item = result if exactly_one else result[0]
        delta = delta or self.delta

        expected = (
            pytest.approx(latitude, abs=delta) if latitude is not ANY else ANY,
            pytest.approx(longitude, abs=delta) if longitude is not ANY else ANY,
            address,
        )
        received = (
            item.latitude,
            item.longitude,
            item.address,
        )
        assert received == expected



import base64
import warnings
from datetime import datetime
from urllib.parse import parse_qs, urlparse

import pytest
from pytz import timezone

from geopy import exc
from geopy.geocoders import GoogleV3
from geopy.point import Point
from test.geocoders.util import BaseTestGeocoder, env


@pytest.mark.skipif(
    not bool(env['GOOGLE_KEY']),
    reason="No GOOGLE_KEY env variable set"
)
class TestGoogleV3(BaseTestGeocoder):
    new_york_point = Point(40.75376406311989, -73.98489005863667)
    america_new_york = timezone("America/New_York")

    @classmethod
    def make_geocoder(cls, **kwargs):
        return GoogleV3(api_key=env.get('GOOGLE_KEY'), **kwargs)

    async def test_user_agent_custom(self):
        geocoder = GoogleV3(
            api_key='mock',
            user_agent='my_user_agent/1.0'
        )
        assert geocoder.headers['User-Agent'] == 'my_user_agent/1.0'

    async def test_configuration_error(self):
        with pytest.raises(exc.ConfigurationError):
            GoogleV3(api_key='mock', client_id='a')
        with pytest.raises(exc.ConfigurationError):
            GoogleV3(api_key='mock', secret_key='a')

    async def test_warning_with_no_api_key(self):
        with warnings.catch_warnings(record=True) as w:
            GoogleV3()
        assert len(w) == 1

    async def test_no_warning_with_no_api_key_but_using_premier(self):
        with warnings.catch_warnings(record=True) as w:
            GoogleV3(client_id='client_id', secret_key='secret_key')
        assert len(w) == 0

    async def test_check_status(self):
        assert self.geocoder._check_status("ZERO_RESULTS") is None
        with pytest.raises(exc.GeocoderQuotaExceeded):
            self.geocoder._check_status("OVER_QUERY_LIMIT")
        with pytest.raises(exc.GeocoderQueryError):
            self.geocoder._check_status("REQUEST_DENIED")
        with pytest.raises(exc.GeocoderQueryError):
            self.geocoder._check_status("INVALID_REQUEST")
        with pytest.raises(exc.GeocoderQueryError):
            self.geocoder._check_status("_")

    async def test_get_signed_url(self):
        geocoder = GoogleV3(
            api_key='mock',
            client_id='my_client_id',
            secret_key=base64.urlsafe_b64encode('my_secret_key'.encode('utf8'))
        )
        assert geocoder.premier
        # the two possible URLs handle both possible orders of the request
        # params; because it's unordered, either is possible, and each has
        # its own hash
        assert geocoder._get_signed_url(
            {'address': '1 5th Ave New York, NY'}
        ) in (
            "https://maps.googleapis.com/maps/api/geocode/json?"
            "address=1+5th+Ave+New+York%2C+NY&client=my_client_id&"
            "signature=Z_1zMBa3Xu0W4VmQfaBR8OQMnDM=",
            "https://maps.googleapis.com/maps/api/geocode/json?"
            "client=my_client_id&address=1+5th+Ave+New+York%2C+NY&"
            "signature=D3PL0cZJrJYfveGSNoGqrrMsz0M="
        )

    async def test_get_signed_url_with_channel(self):
        geocoder = GoogleV3(
            api_key='mock',
            client_id='my_client_id',
            secret_key=base64.urlsafe_b64encode('my_secret_key'.encode('utf8')),
            channel='my_channel'
        )

        signed_url = geocoder._get_signed_url({'address': '1 5th Ave New York, NY'})
        params = parse_qs(urlparse(signed_url).query)

        assert 'channel' in params
        assert 'signature' in params
        assert 'client' in params

    async def test_format_components_param(self):
        f = self.geocoder._format_components_param
        assert f({}) == ''
        assert f([]) == ''

        assert f({'country': 'FR'}) == 'country:FR'
        output = f({'administrative_area': 'CA', 'country': 'FR'})
        # the order the dict is iterated over is not important
        assert output in (
            'administrative_area:CA|country:FR',
            'country:FR|administrative_area:CA'
        ), output

        assert f([('country', 'FR')]) == 'country:FR'
        output = f([
            ('administrative_area', 'CA'),
            ('administrative_area', 'Los Angeles'),
            ('country', 'US')
        ])
        assert (
            output ==
            'administrative_area:CA|administrative_area:Los Angeles|country:US'
        )

        with pytest.raises(ValueError):
            f(None)

        with pytest.raises(ValueError):
            f('administrative_area:CA|country:FR')

    async def test_geocode(self):
        await self.geocode_run(
            {"query": "435 north michigan ave, chicago il 60611 usa"},
            {"latitude": 41.890, "longitude": -87.624},
        )

    async def test_unicode_name(self):
        await self.geocode_run(
            {"query": "\u6545\u5bab"},
            {"latitude": 39.916, "longitude": 116.390},
        )

    async def test_geocode_with_conflicting_components(self):
        await self.geocode_run(
            {
                "query": "santa cruz",
                "components": {
                    "administrative_area": "CA",
                    "country": "FR"
                }
            },
            {},
            expect_failure=True
        )

        await self.geocode_run(
            {
                "query": "santa cruz",
                "components": [
                    ('administrative_area', 'CA'),
                    ('country', 'FR')
                ]
            },
            {},
            expect_failure=True
        )

    async def test_components(self):
        await self.geocode_run(
            {
                "query": "santa cruz",
                "components": {
                    "country": "ES"
                }
            },
            {"latitude": 28.4636296, "longitude": -16.2518467},
        )

        await self.geocode_run(
            {
                "query": "santa cruz",
                "components": [
                    ('country', 'ES')
                ]
            },
            {"latitude": 28.4636296, "longitude": -16.2518467},
        )

    async def test_components_without_query(self):
        await self.geocode_run(
            {
                "components": {"city": "Paris", "country": "FR"},
            },
            {"latitude": 46.227638, "longitude": 2.213749},
        )

        await self.geocode_run(
            {
                "components": [("city", "Paris"), ("country", "FR")],
            },
            {"latitude": 46.227638, "longitude": 2.213749},
        )

    async def test_reverse(self):
        await self.reverse_run(
            {"query": self.new_york_point},
            {"latitude": 40.75376406311989, "longitude": -73.98489005863667},
        )

    async def test_zero_results(self):
        with pytest.raises(exc.GeocoderQueryError):
            await self.geocode_run(
                {"query": ''},
                {},
                expect_failure=True,
            )

    async def test_timezone_datetime(self):
        await self.reverse_timezone_run(
            {"query": self.new_york_point,
             "at_time": datetime.utcfromtimestamp(0)},
            self.america_new_york,
        )

    async def test_timezone_at_time_normalization(self):
        utc_naive_dt = datetime(2010, 1, 1, 0, 0, 0)
        utc_timestamp = 1262304000
        assert (
            utc_timestamp == self.geocoder._normalize_timezone_at_time(utc_naive_dt)
        )

        assert (
            utc_timestamp < self.geocoder._normalize_timezone_at_time(None)
        )

        tz = timezone("Etc/GMT-2")
        local_aware_dt = tz.localize(datetime(2010, 1, 1, 2, 0, 0))
        assert(
            utc_timestamp == self.geocoder._normalize_timezone_at_time(local_aware_dt)
        )

    async def test_timezone_integer_raises(self):
        # In geopy 1.x `at_time` could be an integer -- a unix timestamp.
        # This is an error since geopy 2.0.
        with pytest.raises(exc.GeocoderQueryError):
            await self.reverse_timezone_run(
                {"query": self.new_york_point, "at_time": 0},
                self.america_new_york,
            )

    async def test_timezone_no_date(self):
        await self.reverse_timezone_run(
            {"query": self.new_york_point},
            self.america_new_york,
        )

    async def test_timezone_invalid_at_time(self):
        with pytest.raises(exc.GeocoderQueryError):
            self.geocoder.reverse_timezone(self.new_york_point, at_time="eek")

    async def test_reverse_timezone_unknown(self):
        await self.reverse_timezone_run(
            # Google doesn't return a timezone for Antarctica.
            {"query": "89.0, 1.0"},
            None,
        )

    async def test_geocode_bounds(self):
        await self.geocode_run(
            {"query": "221b Baker St", "bounds": [[50, -2], [55, 2]]},
            {"latitude": 51.52, "longitude": -0.15},
        )

    async def test_geocode_bounds_invalid(self):
        with pytest.raises(exc.GeocoderQueryError):
            await self.geocode_run(
                {"query": "221b Baker St", "bounds": [50, -2, 55]},
                {"latitude": 51.52, "longitude": -0.15},
            )

    async def test_geocode_place_id_invalid(self):
        await self.geocode_run(
            {"place_id": "ChIJOcfP0Iq2j4ARDrXUa7ZWs34"},
            {"latitude": 37.22, "longitude": -122.05}
        )

    async def test_geocode_place_id_not_invalid(self):
        with pytest.raises(exc.GeocoderQueryError):
            await self.geocode_run(
                {"place_id": "xxxxx"},
                {},
                expect_failure=True,
            )

    async def test_place_id_zero_result(self):
        with pytest.raises(exc.GeocoderQueryError):
            await self.geocode_run(
                {"place_id": ""},
                {},
                expect_failure=True,
            )

    async def test_geocode_place_id_with_query(self):
        with pytest.raises(ValueError):
            await self.geocode_run(
                {"place_id": "ChIJOcfP0Iq2j4ARDrXUa7ZWs34",
                 "query": "silicon valley"},
                {}
            )

    async def test_geocode_place_id_with_bounds(self):
        with pytest.raises(ValueError):
            await self.geocode_run(
                {"place_id": "ChIJOcfP0Iq2j4ARDrXUa7ZWs34",
                 "bounds": [50, -2, 55, 2]},
                {}
            )

    async def test_geocode_place_id_with_query_and_bounds(self):
        with pytest.raises(ValueError):
            await self.geocode_run(
                {"place_id": "ChIJOcfP0Iq2j4ARDrXUa7ZWs34",
                 "query": "silicon valley",
                 "bounds": [50, -2, 55, 2]},
                {}
            )



import math
import pickle
import sys
import unittest
import warnings

from geopy.point import Point


class PointTestCase(unittest.TestCase):
    lat = 40.74113
    lon = -73.989656
    alt = 3
    coords = (lat, lon, alt)

    def test_point_float(self):
        point = Point(self.lat, self.lon, self.alt)
        self.assertEqual(point.longitude, self.lon)
        self.assertEqual(point.latitude, self.lat)
        self.assertEqual(point.altitude, self.alt)

    def test_point_str_simple(self):
        for each in ("%s,%s", "%s %s", "%s;%s"):
            point = Point(each % (self.lat, self.lon))
            self.assertEqual(point.longitude, self.lon)
            self.assertEqual(point.latitude, self.lat)
            self.assertEqual(point.altitude, 0)

    def test_point_str_deg(self):
        point = Point("UT: N 39\xb020' 0'' / W 74\xb035' 0''")
        self.assertEqual(point.latitude, 39.333333333333336)
        self.assertEqual(point.longitude, -74.58333333333333)
        self.assertEqual(point.altitude, 0)

    def test_point_format(self):
        point = Point("51 19m 12.9s N, 0 1m 24.95s E")
        self.assertEqual(point.format(), "51 19m 12.9s N, 0 1m 24.95s E")

        point = Point("51 19m 12.9s N, -1 1m 24.95s E, 15000m")
        self.assertEqual(point.format(), "51 19m 12.9s N, 1 1m 24.95s W, 15.0km")

        # TODO
        # point = Point("51 19m 12.9s N, -0 1m 24.95s E")
        # self.assertEqual(point.format(), "51 19m 12.9s N, 0 1m 24.95s W")

        # TODO
        # with self.assertRaises(ValueError):
        #     # Z is not a valid direction
        #     Point("51 19m 12.9s Z, 0 1m 24.95s E")

        with self.assertRaises(ValueError):
            Point("gibberish")

        with self.assertRaises(ValueError):
            # It could be interpreted as `Point(75, 5)`.
            Point("75 5th Avenue, NYC, USA")

    def test_point_from_string(self):
        # Examples are from the docstring of `Point.from_string`.
        self.assertEqual(Point("41.5;-81.0"), (41.5, -81.0, 0.0))
        self.assertEqual(Point("41.5,-81.0"), (41.5, -81.0, 0.0))
        self.assertEqual(Point("41.5 -81.0"), (41.5, -81.0, 0.0))
        self.assertEqual(Point("41.5 N -81.0 W"), (41.5, 81.0, 0.0))
        self.assertEqual(Point("-41.5 S;81.0 E"), (41.5, 81.0, 0.0))
        self.assertEqual(Point("23 26m 22s N 23 27m 30s E"),
                         (23.439444444444444, 23.458333333333332, 0.0))
        self.assertEqual(Point("23 26' 22\" N 23 27' 30\" E"),
                         (23.439444444444444, 23.458333333333332, 0.0))
        self.assertEqual(Point("UT: N 39\xb020' 0'' / W 74\xb035' 0''"),
                         (39.333333333333336, -74.58333333333333, 0.0))

    def test_point_format_altitude(self):
        point = Point(latitude=41.5, longitude=81.0, altitude=2.5)
        self.assertEqual(point.format(), "41 30m 0s N, 81 0m 0s E, 2.5km")
        self.assertEqual(point.format_decimal(), "41.5, 81.0, 2.5km")
        self.assertEqual(point.format_decimal('m'), "41.5, 81.0, 2500.0m")

        point = Point(latitude=41.5, longitude=81.0)
        self.assertEqual(point.format_decimal(), "41.5, 81.0")
        self.assertEqual(point.format_decimal('m'), "41.5, 81.0, 0.0m")

    def test_point_from_iterable(self):
        self.assertEqual(Point(1, 2, 3), Point([1, 2, 3]))
        self.assertEqual(Point(1, 2, 0), Point([1, 2]))
        with self.assertRaises(ValueError):
            Point([1])
        self.assertEqual(Point(0, 0, 0), Point([]))

        with self.assertRaises(ValueError):
            Point([1, 2, 3, 4])

    def test_point_from_single_number(self):
        with self.assertRaises(ValueError):
            # Point from a single number is probably a misuse,
            # thus it's discouraged.
            Point(5)

        # But an explicit zero longitude is fine
        self.assertEqual((5, 0, 0), tuple(Point(5, 0)))

    def test_point_from_point(self):
        point = Point(self.lat, self.lon, self.alt)
        point_copy = Point(point)
        self.assertTrue(point is not point_copy)
        self.assertEqual(tuple(point), tuple(point_copy))

    def test_point_from_generator(self):
        point = Point(i + 10 for i in range(3))
        self.assertEqual((10, 11, 12), tuple(point))
        self.assertEqual((10, 11, 12), tuple(point))

    def test_point_degrees_are_normalized(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            with self.assertRaises(ValueError):
                # Latitude normalization is not allowed
                Point(95, 185, 375)
            self.assertEqual(1, len(w))

            with self.assertRaises(ValueError):
                # Latitude normalization is not allowed
                Point(-95, -185, 375)
            self.assertEqual(2, len(w))

            point = Point(-85, 185, 375)
            self.assertEqual((-85, -175, 375), tuple(point))

            point = Point(85, -185, 375)
            self.assertEqual((85, 175, 375), tuple(point))

            # note that the zeros might be negative
            point = Point(-0.0, -0.0, 375)
            self.assertEqual((0.0, 0.0, 375.0), tuple(point))
            # ensure that negative zeros are normalized to the positive ones
            self.assertEqual((1.0, 1.0, 1.0), tuple(math.copysign(1.0, x) for x in point))

            point = Point(90, 180, 375)
            self.assertEqual((90, 180, 375), tuple(point))
            point = Point(-90, -180, 375)
            self.assertEqual((-90, -180, 375), tuple(point))
            self.assertEqual(2, len(w))

            point = Point(-90, -540, 375)
            self.assertEqual((-90, -180, 375), tuple(point))

            point = Point(90, 540, 375)
            self.assertEqual((90, -180, 375), tuple(point))
            self.assertEqual(2, len(w))

    def test_point_degrees_normalization_does_not_lose_precision(self):
        if sys.float_info.mant_dig != 53:  # pragma: no cover
            raise unittest.SkipTest('This platform does not store floats as '
                                    'IEEE 754 double')
        # IEEE 754 double is stored like this:
        # sign (1 bit) | exponent (11 bit) | fraction (52 bit)
        # \/         \/
        # 0100000111010010011001001001001011000010100000000000000000000000
        #
        # The issue is that there might be a loss in precision during
        # normalization.
        # For example, 180.00000000000003 is stored like this:
        # \/         \/
        # 0100000001100110100000000000000000000000000000000000000000000001
        #
        # And if we do (180.00000000000003 + 180 - 180), then we would get
        # exactly 180.0:
        # 0100000001100110100000000000000000000000000000000000000000000000
        #
        # Notice that the last fraction bit has been lost, because
        # (180.00000000000003 + 180) fraction doesn't fit in 52 bits.
        #
        # This test ensures that such unwanted precision loss is not happening.
        self.assertEqual(tuple(Point(-89.99999999999998, 180.00000000000003)),
                         (-89.99999999999998, -179.99999999999997, 0))
        self.assertEqual(tuple(Point(9.000000000000002, 1.8000000000000003)),
                         (9.000000000000002, 1.8000000000000003, 0))

    def test_unpacking(self):
        point = Point(self.lat, self.lon, self.alt)
        lat, lon, alt = point
        self.assertEqual(lat, self.lat)
        self.assertEqual(lon, self.lon)
        self.assertEqual(alt, self.alt)

    def test_point_no_len(self):
        point = Point(self.lat, self.lon)  # is it 2 or 3?
        with self.assertRaises(TypeError):  # point doesn't support len()
            len(point)

    def test_point_getitem(self):
        point = Point(self.lat, self.lon, self.alt)
        self.assertEqual(point[0], self.lat)
        self.assertEqual(point[1], self.lon)
        self.assertEqual(point[2], self.alt)

    def test_point_slices(self):
        point = Point(self.lat, self.lon, self.alt)
        self.assertEqual((self.lat, self.lon), point[:2])

        self.assertEqual(self.coords, point[:10])
        self.assertEqual(self.coords, point[:])
        self.assertEqual(self.coords[::-1], point[::-1])

        with self.assertRaises(IndexError):
            point[10]

        with self.assertRaises(TypeError):
            point[None]

        point[0:2] = (self.lat + 10, self.lon + 10)
        self.assertEqual((self.lat + 10, self.lon + 10, self.alt),
                         tuple(point))

    def test_point_setitem(self):
        point = Point(self.lat + 10, self.lon + 10, self.alt + 10)
        for each in (0, 1, 2):
            point[each] = point[each] - 10

        self.assertEqual(point[0], self.lat)
        self.assertEqual(point[1], self.lon)
        self.assertEqual(point[2], self.alt)

        self.assertEqual(self.coords, tuple(point))
        self.assertEqual(point.latitude, self.lat)
        self.assertEqual(point.longitude, self.lon)
        self.assertEqual(point.altitude, self.alt)

    def test_point_setitem_normalization(self):
        point = Point()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            with self.assertRaises(ValueError):
                point[0] = 100
            self.assertEqual(1, len(w))
            self.assertEqual((0, 0, 0), tuple(point))
            point[0] = -80
            point[1] = 200
            # Please note that attribute assignments are not normalized.
            # Only __setitem__ assignments are.
            self.assertEqual((-80, -160, 0), tuple(point))
            with self.assertRaises(ValueError):
                point[1] = float("nan")
            self.assertEqual(1, len(w))
            self.assertEqual((-80, -160, 0), tuple(point))

    def test_point_assign_coordinates(self):
        point = Point(self.lat + 10, self.lon + 10, self.alt + 10)
        point.latitude = self.lat
        point.longitude = self.lon
        point.altitude = self.alt

        self.assertEqual(point[0], self.lat)
        self.assertEqual(point[1], self.lon)
        self.assertEqual(point[2], self.alt)

        self.assertEqual(self.coords, tuple(point))
        self.assertEqual(point.latitude, self.lat)
        self.assertEqual(point.longitude, self.lon)
        self.assertEqual(point.altitude, self.alt)

    def test_point_eq(self):
        self.assertEqual(
            Point(self.lat, self.lon),
            Point("%s %s" % (self.lat, self.lon))
        )

    def test_point_ne(self):
        self.assertTrue(
            Point(self.lat, self.lon, self.alt) !=
            Point(self.lat+10, self.lon-10, self.alt)
        )

    def test_point_comparison_does_not_raise_exceptions(self):
        point = Point(self.lat, self.lon, self.alt)
        number = 1
        not_iterable = object()
        self.assertFalse(point == number)
        self.assertTrue(point != number)
        self.assertFalse(point == not_iterable)
        self.assertTrue(point != not_iterable)

    def test_point_comparison_with_empty_values(self):
        empty_values = (None, '', [], (), {})  # Django validators
        point_nonempty = Point(self.lat, self.lon, self.alt)
        point_empty = Point()  # actually Point(0, 0, 0), which is not "empty"
        self.assertFalse(point_nonempty in empty_values)
        self.assertTrue(point_nonempty)

        # Point() == (0, 0, 0).
        self.assertEqual(Point(), (0, 0, 0))
        # Currently Point can't distinguish between zero and unset coordinate
        # values, so we cannot tell if `point_empty` is really "empty"
        # (i.e. unset, like `Point()`) or is just a point at the center
        # (which is obviously not "empty"). That's why on the next line
        # we assume that `point_empty` is not in the `empty_values` list.
        self.assertFalse(point_empty in empty_values)
        # bool(Point(0, 0, 0)) should also be true
        self.assertTrue(point_empty)

    def test_point_comparison_respects_lists(self):
        point = Point(self.lat, self.lon, self.alt)
        l_eq = [self.lat, self.lon, self.alt]
        l_ne = [self.lat + 1, self.lon, self.alt]
        self.assertTrue(l_eq == point)
        self.assertTrue(point == l_eq)
        self.assertFalse(l_eq != point)
        self.assertFalse(point != l_eq)

        self.assertFalse(l_ne == point)
        self.assertFalse(point == l_ne)
        self.assertTrue(l_ne != point)
        self.assertTrue(point != l_ne)

    def test_point_comparison_ignores_strings(self):
        point = Point("1", "2", "3")
        self.assertFalse(point == "123")
        self.assertTrue(point != "123")
        self.assertTrue(point == (1, 2, 3))

    def test_point_is_picklable(self):
        point = Point(self.lat, self.lon, self.alt)
        # https://docs.python.org/2/library/pickle.html#data-stream-format
        for protocol in (0, 1, 2, -1):
            pickled = pickle.dumps(point, protocol=protocol)
            point_unp = pickle.loads(pickled)
            self.assertEqual(point, point_unp)
            self.assertEqual(self.coords, point_unp)



import pytest

from geopy.geocoders import GeocodeEarth
from test.geocoders.pelias import BaseTestPelias
from test.geocoders.util import env


@pytest.mark.skipif(
    not bool(env['GEOCODEEARTH_KEY']),
    reason="No GEOCODEEARTH_KEY env variable set"
)
class TestGeocodeEarth(BaseTestPelias):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return GeocodeEarth(env['GEOCODEEARTH_KEY'],
                            **kwargs)



from distutils.version import LooseVersion

from geopy import __version__, __version_info__, get_version


def test_version():
    assert isinstance(__version__, str) and __version__


def test_version_info():
    expected_version_info = tuple(LooseVersion(__version__).version)
    assert expected_version_info == __version_info__


def test_get_version():
    version = get_version()
    assert isinstance(version, str) and version == __version__



import pytest

from geopy.geocoders import AzureMaps
from test.geocoders.tomtom import BaseTestTomTom
from test.geocoders.util import env


@pytest.mark.skipif(
    not bool(env.get('AZURE_SUBSCRIPTION_KEY')),
    reason="No AZURE_SUBSCRIPTION_KEY env variable set"
)
class TestAzureMaps(BaseTestTomTom):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return AzureMaps(env['AZURE_SUBSCRIPTION_KEY'], timeout=3,
                         **kwargs)



import math
import unittest
import warnings

from geopy.distance import (
    EARTH_RADIUS,
    Distance,
    GeodesicDistance,
    GreatCircleDistance,
    distance,
    lonlat,
)
from geopy.point import Point

EARTH_CIRCUMFERENCE = 2 * math.pi * EARTH_RADIUS
NORTH_POLE = Point(90, 0)
SOUTH_POLE = Point(-90, 0)
FIJI = Point(-16.1333333, 180.0)  # Vunikondi, Fiji


class CommonDistanceComputationCases:

    cls = None

    def test_zero_measure(self):
        self.cls(
            (40.753152999999998, -73.982275999999999),
            (40.753152999999998, -73.982275999999999)
        )

    def test_should_have_length_when_only_given_length(self):
        distance = 1
        self.assertEqual(self.cls(distance).kilometers, distance)

    def test_should_have_zero_distance_for_coincident_points(self):
        self.assertEqual(self.cls((0, 0), (0, 0)).kilometers, 0)

    def test_should_have_nonzero_distance_for_distinct_points(self):
        self.assertTrue(self.cls((0, 0), (0, 1)).kilometers > 0)

    def test_max_longitude(self):
        distance = self.cls(kilometers=1.0)
        destination = distance.destination(FIJI, 45)
        self.assertAlmostEqual(destination.longitude, -179.99338, 4)

    def test_should_compute_distance_for_trip_between_poles(self):
        distance = self.cls(SOUTH_POLE, NORTH_POLE)
        expected_distance = EARTH_CIRCUMFERENCE / 2
        self.assertAlmostEqual(distance.kilometers, expected_distance, -2)

    def test_should_compute_destination_for_trip_between_poles(self):
        distance = self.cls(EARTH_CIRCUMFERENCE / 2)
        destination = distance.destination(NORTH_POLE, 0)
        self.assertAlmostEqual(destination.latitude, -90, 0)
        self.assertAlmostEqual(destination.longitude, 0)

    def test_should_recognize_equivalence_of_pos_and_neg_180_longitude(self):
        distance1 = self.cls((0, -180), (0, 180)).kilometers
        distance2 = self.cls((0, 180), (0, -180)).kilometers
        self.assertAlmostEqual(distance1, 0)
        self.assertAlmostEqual(distance2, 0)

    def test_should_compute_distance_across_antimeridian(self):
        nines = 1 - 1e-30  # 0.(9)
        distance = self.cls((0, -179 - nines),
                            (0, 179 + nines)).kilometers
        self.assertAlmostEqual(distance, 0)

    def test_should_compute_destination_across_antimeridian(self):
        nines = 1 - 1e-30  # 0.(9)
        distance = self.cls(10)
        point = distance.destination((0, -179 - nines), -90)
        self.assertAlmostEqual(point.latitude, 0.0)
        self.assertAlmostEqual(point.longitude, 179.91, 3)

    def test_should_not_tolerate_nans(self):
        nan = float('nan')
        with self.assertRaises(ValueError):
            self.cls((nan, nan), (1, 1)).kilometers
        with self.assertRaises(ValueError):
            self.cls((nan, 1), (1, nan)).kilometers
        with self.assertRaises(ValueError):
            self.cls((nan, 1), (nan, 1)).kilometers

    def test_should_compute_distance_for_multiple_points_pairwise(self):
        dist_total = self.cls((10, 20), (40, 60), (0, 80), (0, 10))
        dist1 = self.cls((10, 20), (40, 60))
        dist2 = self.cls((40, 60), (0, 80))
        dist3 = self.cls((0, 80), (0, 10))

        self.assertAlmostEqual(dist_total.kilometers,
                               dist1.km + dist2.km + dist3.km)

    def test_should_raise_when_using_single_numbers_as_points(self):
        # Each argument is expected to be a Point. If it's not a point,
        # it will be wrapped in Point.
        # Point(10) equals to Point(10, 0).
        with self.assertRaises(ValueError):
            self.cls(10, 20)

        # no error: explicit tuples are not that suspicious
        self.cls((10, 0), (20, 0))

    def test_should_tolerate_tuples_with_textual_numbers(self):
        dist1 = self.cls(("1", "30"), ("20", "60"))
        dist2 = self.cls((1, 30), (20, 60))
        self.assertAlmostEqual(dist1.kilometers, dist2.kilometers)

    def test_should_warn_for_mixed_up_lat_lon(self):
        lat = 40
        lon = 120  # should exceed max lat (abs 90) to cause a warning
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            # The correct order is (lat, lon).
            with self.assertRaises(ValueError):
                self.cls((lon, lat), (lon - 10, lat))
            self.assertEqual(1, len(w))


class CommonMathematicalOperatorCases:

    cls = None

    def test_should_be_able_to_add_distances(self):
        added = self.cls(1.0) + self.cls(1.0)
        self.assertAlmostEqual(added.kilometers, 2.0)

    def test_should_not_allow_adding_with_objects_that_arent_distances(self):
        with self.assertRaises(TypeError):
            self.cls(1.0) + 5

    def test_should_be_able_to_negate_distances(self):
        distance = self.cls(1.0)
        self.assertAlmostEqual(-(distance.kilometers),
                               (-distance).kilometers)

    def test_should_be_able_to_subtract_distances(self):
        subtracted = self.cls(2.0) - self.cls(1.0)
        self.assertAlmostEqual(subtracted.kilometers, 1)

    def test_should_be_able_to_multiply_distances_by_floats(self):
        self.assertAlmostEqual((self.cls(2.0) * 2.0).kilometers, 4.0)

    def test_should_not_be_able_to_multiply_distances_by_distances(self):
        with self.assertRaises(TypeError):
            self.cls(1.0) * self.cls(2.0)

    def test_should_be_able_to_divide_distances_by_distances(self):
        ratio = self.cls(4.0) / self.cls(2.0)
        self.assertIsInstance(ratio, float)
        self.assertAlmostEqual(ratio, 2.0)

    def test_should_be_able_to_divide_distances_by_floats(self):
        divided_distance = self.cls(4.0) / 2.0
        self.assertIsInstance(divided_distance, self.cls)
        self.assertAlmostEqual(divided_distance.kilometers, 2.0)

    def test_should_be_able_to_take_absolute_value_of_distances(self):
        self.assertAlmostEqual(abs(self.cls(-1.0)).kilometers, 1.0)

    def test_should_be_true_in_boolean_context_when_nonzero_length(self):
        self.assertTrue(self.cls(1.0))

    def test_should_be_false_in_boolean_context_when_zero_length(self):
        self.assertFalse(self.cls(0))

    def test_should_get_consistent_results_for_distance_calculations(self):
        distance1, distance2 = [self.cls((0, 0), (0, 1))
                                for _ in range(2)]
        self.assertEqual(distance1.kilometers, distance2.kilometers)


class CommonConversionCases:

    cls = None

    def test_should_convert_to_kilometers(self):
        self.assertEqual(self.cls(1.0).kilometers, 1.0)

    def test_should_convert_to_kilometers_with_abbreviation(self):
        self.assertEqual(self.cls(1.0).km, 1.0)

    def test_should_convert_to_meters(self):
        self.assertEqual(self.cls(1.0).meters, 1000.0)

    def test_should_convert_to_meters_with_abbreviation(self):
        self.assertEqual(self.cls(1.0).m, 1000.0)

    def test_should_convert_to_miles(self):
        self.assertAlmostEqual(self.cls(1.0).miles, 0.621371192)

    def test_should_convert_to_miles_with_abbreviation(self):
        self.assertAlmostEqual(self.cls(1.0).mi, 0.621371192)

    def test_should_convert_to_feet(self):
        self.assertAlmostEqual(self.cls(1.0).feet, 3280.8399, 4)

    def test_should_convert_to_feet_with_abbreviation(self):
        self.assertAlmostEqual(self.cls(1.0).ft, 3280.8399, 4)

    def test_should_convert_to_nautical_miles(self):
        self.assertAlmostEqual(self.cls(1.0).nautical, 0.539956803)

    def test_should_convert_to_nautical_miles_with_abbrevation(self):
        self.assertAlmostEqual(self.cls(1.0).nm, 0.539956803)

    def test_should_convert_from_meters(self):
        self.assertEqual(self.cls(meters=1.0).km, 0.001)

    def test_should_convert_from_feet(self):
        self.assertAlmostEqual(self.cls(feet=1.0).km, 0.0003048, 8)

    def test_should_convert_from_miles(self):
        self.assertAlmostEqual(self.cls(miles=1.0).km, 1.6093440)

    def test_should_convert_from_nautical_miles(self):
        self.assertAlmostEqual(self.cls(nautical=1.0).km, 1.8520000)


class CommonComparisonCases:

    cls = None

    def test_should_support_comparison_with_distance(self):
        self.assertTrue(self.cls(1) <= self.cls(1))
        self.assertTrue(self.cls(1) >= self.cls(1))
        self.assertTrue(self.cls(1) < self.cls(2))
        self.assertTrue(self.cls(2) > self.cls(1))
        self.assertTrue(self.cls(1) == self.cls(1))
        self.assertTrue(self.cls(1) != self.cls(2))

        self.assertFalse(self.cls(2) <= self.cls(1))
        self.assertFalse(self.cls(1) >= self.cls(2))
        self.assertFalse(self.cls(2) < self.cls(1))
        self.assertFalse(self.cls(1) > self.cls(2))
        self.assertFalse(self.cls(1) == self.cls(2))
        self.assertFalse(self.cls(1) != self.cls(1))

    def test_should_support_comparison_with_number(self):
        self.assertTrue(1 <= self.cls(1))
        self.assertTrue(1 >= self.cls(1))
        self.assertTrue(1 < self.cls(2))
        self.assertTrue(2 > self.cls(1))
        self.assertTrue(1 == self.cls(1))
        self.assertTrue(1 != self.cls(2))

        self.assertFalse(2 <= self.cls(1))
        self.assertFalse(1 >= self.cls(2))
        self.assertFalse(2 < self.cls(1))
        self.assertFalse(1 > self.cls(2))
        self.assertFalse(1 == self.cls(2))
        self.assertFalse(1 != self.cls(1))


class CommonDistanceCases(CommonDistanceComputationCases,
                          CommonMathematicalOperatorCases,
                          CommonConversionCases,
                          CommonComparisonCases):
    pass


class TestWhenInstantiatingBaseDistanceClass(unittest.TestCase):
    def test_should_not_be_able_to_instantiate(self):
        with self.assertRaises(TypeError):
            Distance((0, 0), (0, 180))


class TestDefaultDistanceClass(unittest.TestCase):

    def test_default_distance(self):
        self.assertEqual(distance(132).km, 132)

    def test_lonlat_function(self):
        newport_ri_xy = (-71.312796, 41.49008)
        point = lonlat(*newport_ri_xy)
        self.assertEqual(point, (41.49008, -71.312796, 0))


class TestWhenComputingGreatCircleDistance(CommonDistanceCases,
                                           unittest.TestCase):
    cls = GreatCircleDistance

    def test_should_compute_distance_for_half_trip_around_equator(self):
        distance_around_earth = self.cls((0, 0), (0, 180)).kilometers
        self.assertEqual(distance_around_earth, EARTH_CIRCUMFERENCE / 2)

    def test_should_compute_destination_for_half_trip_around_equator(self):
        distance = self.cls(EARTH_CIRCUMFERENCE / 2)
        destination = distance.destination((0, 0), 0)
        self.assertAlmostEqual(destination.latitude, 0)
        self.assertAlmostEqual(destination.longitude, 180)

    def test_different_altitudes_error(self):
        with self.assertRaises(ValueError):
            # Different altitudes raise an exception:
            self.cls((10, 10, 10), (20, 20, 15))

        # Equal non-zero altitudes don't raise:
        self.cls((10, 10, 10), (20, 20, 10))


class TestWhenComputingGeodesicDistance(CommonDistanceCases,
                                        unittest.TestCase):

    cls = GeodesicDistance

    def setUp(self):
        self.original_ellipsoid = self.cls.ELLIPSOID

    def tearDown(self):
        self.cls.ELLIPSOID = self.original_ellipsoid

    def test_different_altitudes_error(self):
        with self.assertRaises(ValueError):
            # Different altitudes raise an exception:
            self.cls((10, 10, 10), (20, 20, 15))

        # Equal non-zero altitudes don't raise:
        self.cls((10, 10, 10), (20, 20, 10))

    def test_miscellaneous_high_accuracy_cases(self):

        testcases = [
          [35.60777, -139.44815, 111.098748429560326,
           -11.17491, -69.95921, 129.289270889708762,
           8935244.5604818305],
          [55.52454, 106.05087, 22.020059880982801,
           77.03196, 197.18234, 109.112041110671519,
           4105086.1713924406],
          [-21.97856, 142.59065, -32.44456876433189,
           41.84138, 98.56635, -41.84359951440466,
           8394328.894657671],
          [-66.99028, 112.2363, 173.73491240878403,
           -12.70631, 285.90344, 2.512956620913668,
           11150344.2312080241],
          [-17.42761, 173.34268, -159.033557661192928,
           -15.84784, 5.93557, -20.787484651536988,
           16076603.1631180673],
          [32.84994, 48.28919, 150.492927788121982,
           -56.28556, 202.29132, 48.113449399816759,
           16727068.9438164461],
          [6.96833, 52.74123, 92.581585386317712,
           -7.39675, 206.17291, 90.721692165923907,
           17102477.2496958388],
          [-50.56724, -16.30485, -105.439679907590164,
           -33.56571, -94.97412, -47.348547835650331,
           6455670.5118668696],
          [-58.93002, -8.90775, 140.965397902500679,
           -8.91104, 133.13503, 19.255429433416599,
           11756066.0219864627],
          [-68.82867, -74.28391, 93.774347763114881,
           -50.63005, -8.36685, 34.65564085411343,
           3956936.926063544],
          [-10.62672, -32.0898, -86.426713286747751,
           5.883, -134.31681, -80.473780971034875,
           11470869.3864563009],
          [-21.76221, 166.90563, 29.319421206936428,
           48.72884, 213.97627, 43.508671946410168,
           9098627.3986554915],
          [-19.79938, -174.47484, 71.167275780171533,
           -11.99349, -154.35109, 65.589099775199228,
           2319004.8601169389],
          [-11.95887, -116.94513, 92.712619830452549,
           4.57352, 7.16501, 78.64960934409585,
           13834722.5801401374],
          [-87.85331, 85.66836, -65.120313040242748,
           66.48646, 16.09921, -4.888658719272296,
           17286615.3147144645],
          [1.74708, 128.32011, -101.584843631173858,
           -11.16617, 11.87109, -86.325793296437476,
           12942901.1241347408],
          [-25.72959, -144.90758, -153.647468693117198,
           -57.70581, -269.17879, -48.343983158876487,
           9413446.7452453107],
          [-41.22777, 122.32875, 14.285113402275739,
           -7.57291, 130.37946, 10.805303085187369,
           3812686.035106021],
          [11.01307, 138.25278, 79.43682622782374,
           6.62726, 247.05981, 103.708090215522657,
           11911190.819018408],
          [-29.47124, 95.14681, -163.779130441688382,
           -27.46601, -69.15955, -15.909335945554969,
           13487015.8381145492]]
        d = self.cls(ellipsoid='WGS-84')
        km = 1000
        for l in testcases:
            (lat1, lon1, azi1, lat2, lon2, azi2, s12) = l
            p1, p2 = Point(lat1, lon1), Point(lat2, lon2)
            s12a = d.measure(p1, p2) * km
            self.assertAlmostEqual(s12a, s12, delta=1e-8)
            p = d.destination(p1, azi1, s12/km)
            self.assertAlmostEqual(p.latitude, p2.latitude, delta=1e-13)
            self.assertAlmostEqual(p.longitude, p2.longitude, delta=1e-12)
            p = d.destination(p2, azi2, -s12/km)
            self.assertAlmostEqual(p.latitude, p1.latitude, delta=1e-13)
            self.assertAlmostEqual(p.longitude, p1.longitude, delta=1e-12)



from geopy.geocoders.base import DEFAULT_SENTINEL
from geopy.geocoders.pelias import Pelias

__all__ = ("GeocodeEarth", )


class GeocodeEarth(Pelias):
    """geocode.earth, a Pelias-based service provided by the developers
    of Pelias itself.
    """

    def __init__(
            self,
            api_key,
            *,
            domain='api.geocode.earth',
            timeout=DEFAULT_SENTINEL,
            proxies=DEFAULT_SENTINEL,
            user_agent=None,
            scheme=None,
            ssl_context=DEFAULT_SENTINEL,
            adapter_factory=None
    ):
        """
        :param str api_key: Geocode.earth API key, required.

        :param str domain: Specify a custom domain for Pelias API.

        :param int timeout:
            See :attr:`geopy.geocoders.options.default_timeout`.

        :param dict proxies:
            See :attr:`geopy.geocoders.options.default_proxies`.

        :param str user_agent:
            See :attr:`geopy.geocoders.options.default_user_agent`.

        :param str scheme:
            See :attr:`geopy.geocoders.options.default_scheme`.

        :type ssl_context: :class:`ssl.SSLContext`
        :param ssl_context:
            See :attr:`geopy.geocoders.options.default_ssl_context`.

        :param callable adapter_factory:
            See :attr:`geopy.geocoders.options.default_adapter_factory`.

            .. versionadded:: 2.0

        """
        super().__init__(
            api_key=api_key,
            domain=domain,
            timeout=timeout,
            proxies=proxies,
            user_agent=user_agent,
            scheme=scheme,
            ssl_context=ssl_context,
            adapter_factory=adapter_factory,
        )



from functools import partial
from urllib.parse import urlencode

from geopy.exc import GeocoderQueryError, GeocoderQuotaExceeded
from geopy.geocoders.base import DEFAULT_SENTINEL, Geocoder
from geopy.location import Location
from geopy.util import logger

__all__ = ("OpenCage", )


class OpenCage(Geocoder):
    """Geocoder using the OpenCageData API.

    Documentation at:
        https://opencagedata.com/api
    """

    api_path = '/geocode/v1/json'

    def __init__(
            self,
            api_key,
            *,
            domain='api.opencagedata.com',
            scheme=None,
            timeout=DEFAULT_SENTINEL,
            proxies=DEFAULT_SENTINEL,
            user_agent=None,
            ssl_context=DEFAULT_SENTINEL,
            adapter_factory=None
    ):
        """

        :param str api_key: The API key required by OpenCageData
            to perform geocoding requests. You can get your key here:
            https://opencagedata.com/

        :param str domain: Currently it is ``'api.opencagedata.com'``, can
            be changed for testing purposes.

        :param str scheme:
            See :attr:`geopy.geocoders.options.default_scheme`.

        :param int timeout:
            See :attr:`geopy.geocoders.options.default_timeout`.

        :param dict proxies:
            See :attr:`geopy.geocoders.options.default_proxies`.

        :param str user_agent:
            See :attr:`geopy.geocoders.options.default_user_agent`.

        :type ssl_context: :class:`ssl.SSLContext`
        :param ssl_context:
            See :attr:`geopy.geocoders.options.default_ssl_context`.

        :param callable adapter_factory:
            See :attr:`geopy.geocoders.options.default_adapter_factory`.

            .. versionadded:: 2.0
        """
        super().__init__(
            scheme=scheme,
            timeout=timeout,
            proxies=proxies,
            user_agent=user_agent,
            ssl_context=ssl_context,
            adapter_factory=adapter_factory,
        )

        self.api_key = api_key
        self.domain = domain.strip('/')
        self.api = '%s://%s%s' % (self.scheme, self.domain, self.api_path)

    def geocode(
            self,
            query,
            *,
            bounds=None,
            country=None,
            language=None,
            exactly_one=True,
            timeout=DEFAULT_SENTINEL
    ):
        """
        Return a location point by address.

        :param str query: The address or query you wish to geocode.

        :param str language: an IETF format language code (such as `es`
            for Spanish or pt-BR for Brazilian Portuguese); if this is
            omitted a code of `en` (English) will be assumed by the remote
            service.

        :type bounds: list or tuple of 2 items of :class:`geopy.point.Point` or
            ``(latitude, longitude)`` or ``"%(latitude)s, %(longitude)s"``.
        :param bounds: Provides the geocoder with a hint to the region
            that the query resides in. This value will help the geocoder
            but will not restrict the possible results to the supplied
            region. The bounds parameter should be specified as 2
            coordinate points -- corners of a bounding box.
            Example: ``[Point(22, 180), Point(-22, -180)]``.

        :param country: Restricts the results to the specified
            country or countries. The country code is a 2 character code as
            defined by the ISO 3166-1 Alpha 2 standard (e.g. ``fr``).
            Might be a Python list of strings.
        :type country: str or list

        :param bool exactly_one: Return one result or a list of results, if
            available.

        :param int timeout: Time, in seconds, to wait for the geocoding service
            to respond before raising a :class:`geopy.exc.GeocoderTimedOut`
            exception. Set this only if you wish to override, on this call
            only, the value set during the geocoder's initialization.

        :rtype: ``None``, :class:`geopy.location.Location` or a list of them, if
            ``exactly_one=False``.

        """
        params = {
            'key': self.api_key,
            'q': query,
        }
        if bounds:
            params['bounds'] = self._format_bounding_box(
                bounds, "%(lon1)s,%(lat1)s,%(lon2)s,%(lat2)s")
        if language:
            params['language'] = language

        if not country:
            country = []
        if isinstance(country, str):
            country = [country]
        if country:
            params['countrycode'] = ",".join(country)

        url = "?".join((self.api, urlencode(params)))

        logger.debug("%s.geocode: %s", self.__class__.__name__, url)
        callback = partial(self._parse_json, exactly_one=exactly_one)
        return self._call_geocoder(url, callback, timeout=timeout)

    def reverse(
            self,
            query,
            *,
            language=None,
            exactly_one=True,
            timeout=DEFAULT_SENTINEL
    ):
        """
        Return an address by location point.

        :param query: The coordinates for which you wish to obtain the
            closest human-readable addresses.
        :type query: :class:`geopy.point.Point`, list or tuple of ``(latitude,
            longitude)``, or string as ``"%(latitude)s, %(longitude)s"``.

        :param str language: The language in which to return results.

        :param bool exactly_one: Return one result or a list of results, if
            available.

        :param int timeout: Time, in seconds, to wait for the geocoding service
            to respond before raising a :class:`geopy.exc.GeocoderTimedOut`
            exception. Set this only if you wish to override, on this call
            only, the value set during the geocoder's initialization.

        :rtype: ``None``, :class:`geopy.location.Location` or a list of them, if
            ``exactly_one=False``.

        """

        params = {
            'key': self.api_key,
            'q': self._coerce_point_to_string(query),
        }
        if language:
            params['language'] = language

        url = "?".join((self.api, urlencode(params)))
        logger.debug("%s.reverse: %s", self.__class__.__name__, url)
        callback = partial(self._parse_json, exactly_one=exactly_one)
        return self._call_geocoder(url, callback, timeout=timeout)

    def _parse_json(self, page, exactly_one=True):
        '''Returns location, (latitude, longitude) from json feed.'''

        places = page.get('results', [])
        if not len(places):
            self._check_status(page.get('status'))
            return None

        def parse_place(place):
            '''Get the location, lat, lng from a single json place.'''
            location = place.get('formatted')
            latitude = place['geometry']['lat']
            longitude = place['geometry']['lng']
            return Location(location, (latitude, longitude), place)

        if exactly_one:
            return parse_place(places[0])
        else:
            return [parse_place(place) for place in places]

    def _check_status(self, status):
        """
        Validates error statuses.
        """
        status_code = status['code']
        if status_code == 429:
            # Rate limit exceeded
            raise GeocoderQuotaExceeded(
                'The given key has gone over the requests limit in the 24'
                ' hour period or has submitted too many requests in too'
                ' short a period of time.'
            )
        if status_code == 200:
            # When there are no results, just return.
            return

        if status_code == 403:
            raise GeocoderQueryError(
                'Your request was denied.'
            )
        else:
            raise GeocoderQueryError('Unknown error.')



import asyncio
import atexit
import contextlib
import importlib
import inspect
import os
from collections import defaultdict
from functools import partial
from statistics import mean, median
from time import sleep
from timeit import default_timer
from unittest.mock import patch
from urllib.parse import urlparse

import pytest

import geopy.geocoders
from geopy.adapters import AdapterHTTPError, BaseAsyncAdapter, BaseSyncAdapter
from geopy.geocoders.base import _DEFAULT_ADAPTER_CLASS

# pytest-aiohttp calls `inspect.isasyncgenfunction` to detect
# async generators in fixtures.
# To support Python 3.5 we use `async_generator` library.
# However:
# - Since Python 3.6 there is a native implementation of
#   `inspect.isasyncgenfunction`, but it returns False
#   for `async_generator`'s functions.
# - The stock `async_generator.isasyncgenfunction` doesn't detect
#   generators wrapped in `@pytest.fixture`.
#
# Thus we resort to monkey-patching it (for now).
if getattr(inspect, "isasyncgenfunction", None) is not None:
    # >=py36
    original_isasyncgenfunction = inspect.isasyncgenfunction
else:
    # ==py35
    original_isasyncgenfunction = lambda func: False  # noqa


def isasyncgenfunction(obj):
    if original_isasyncgenfunction(obj):
        return True
    # Detect async_generator function, possibly wrapped in `@pytest.fixture`:
    # See https://github.com/python-trio/async_generator/blob/v1.10/async_generator/_impl.py#L451-L455  # noqa
    return bool(getattr(obj, "_async_gen_function", None))


inspect.isasyncgenfunction = isasyncgenfunction


def load_adapter_cls(adapter_ref):
    actual_adapter_class = _DEFAULT_ADAPTER_CLASS
    if adapter_ref:
        module_s, cls_s = adapter_ref.rsplit(".", 1)
        module = importlib.import_module(module_s)
        actual_adapter_class = getattr(module, cls_s)
    return actual_adapter_class


max_retries = int(os.getenv('GEOPY_TEST_RETRIES', 2))
error_wait_seconds = float(os.getenv('GEOPY_TEST_ERROR_WAIT_SECONDS', 3))
no_retries_for_hosts = set(os.getenv('GEOPY_TEST_NO_RETRIES_FOR_HOSTS', '').split(','))
default_adapter = load_adapter_cls(os.getenv('GEOPY_TEST_ADAPTER'))
default_adapter_is_async = issubclass(default_adapter, BaseAsyncAdapter)
retry_status_codes = (429,)


def pytest_report_header(config):
    internet_access = "allowed" if _is_internet_access_allowed(config) else "disabled"
    adapter_type = "async" if default_adapter_is_async else "sync"
    return (
        "geopy:\n"
        "    internet access: %s\n"
        "    adapter: %r\n"
        "    adapter type: %s"
        % (internet_access, default_adapter, adapter_type)
    )


def pytest_addoption(parser):
    # This option will probably be used in downstream packages,
    # thus it should be considered a public interface.
    parser.addoption(
        "--skip-tests-requiring-internet",
        action="store_true",
        help="Skip tests requiring Internet access.",
    )


def _is_internet_access_allowed(config):
    return not config.getoption("--skip-tests-requiring-internet")


@pytest.fixture(scope='session')
def is_internet_access_allowed(request):
    return _is_internet_access_allowed(request.config)


@pytest.fixture
def skip_if_internet_access_is_not_allowed(is_internet_access_allowed):
    # Used in test_adapters.py, which doesn't use the injected adapter below.
    if not is_internet_access_allowed:
        pytest.skip("Skipping a test requiring Internet access")


@pytest.fixture(autouse=True, scope="session")
def loop():
    # Geocoder instances have class scope, so the event loop
    # should have session scope.
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    return loop


def netloc_from_url(url):
    return urlparse(url).netloc


def pretty_dict_format(heading, dict_to_format,
                       item_prefix='  ', legend='',
                       value_mapper=lambda v: v):
    s = [heading]
    if not dict_to_format:
        s.append(item_prefix + '-- empty --')
    else:
        max_key_len = max(len(k) for k in dict_to_format.keys())
        for k, v in sorted(dict_to_format.items()):
            s.append('%s%s%s' % (item_prefix, k.ljust(max_key_len + 2),
                                 value_mapper(v)))
        if legend:
            s.append('')
            s.append('* %s' % legend)
    s.append('')  # trailing newline
    return '\n'.join(s)


class RequestsMonitor:
    """RequestsMonitor holds statistics of Adapter requests."""

    def __init__(self):
        self.host_stats = defaultdict(lambda: dict(count=0, retries=0, times=[]))

    def record_request(self, url):
        hostname = netloc_from_url(url)
        self.host_stats[hostname]['count'] += 1

    def record_retry(self, url):
        hostname = netloc_from_url(url)
        self.host_stats[hostname]['retries'] += 1

    @contextlib.contextmanager
    def record_response(self, url):
        start = default_timer()
        try:
            yield
        finally:
            end = default_timer()
            hostname = netloc_from_url(url)
            self.host_stats[hostname]['times'].append(end - start)

    def __str__(self):
        def value_mapper(v):
            tv = v['times']
            times_format = (
                "min:%5.2fs, median:%5.2fs, max:%5.2fs, mean:%5.2fs, total:%5.2fs"
            )
            if tv:
                # min/max require a non-empty sequence.
                times = times_format % (min(tv), median(tv), max(tv), mean(tv), sum(tv))
            else:
                nan = float("nan")
                times = times_format % (nan, nan, nan, nan, 0)

            count = "count:%3d" % v['count']
            retries = "retries:%3d" % v['retries'] if v['retries'] else ""
            return "; ".join(s for s in (count, times, retries) if s)

        legend = (
            "count – number of requests (excluding retries); "
            "min, median, max, mean, total – request duration statistics "
            "(excluding failed requests); retries – number of retries."
        )
        return pretty_dict_format('Request statistics per hostname',
                                  self.host_stats,
                                  legend=legend,
                                  value_mapper=value_mapper)


@pytest.fixture(scope='session')
def requests_monitor():
    return RequestsMonitor()


@pytest.fixture(autouse=True, scope='session')
def print_requests_monitor_report(requests_monitor):
    yield

    def report():
        print(str(requests_monitor))

    # https://github.com/pytest-dev/pytest/issues/2704
    # https://stackoverflow.com/a/38806934
    atexit.register(report)


@pytest.fixture(autouse=True, scope='session')
def patch_adapter(requests_monitor, is_internet_access_allowed):
    """
    Patch the default Adapter to provide the following features:
        - Retry failed requests. Makes test runs more stable.
        - Track statistics with RequestsMonitor.
        - Skip tests requiring Internet access when Internet access is not allowed.
    """

    if default_adapter_is_async:

        class AdapterProxy(BaseAdapterProxy, BaseAsyncAdapter):
            async def __aenter__(self):
                assert await self.adapter.__aenter__() is self.adapter
                return self

            async def __aexit__(self, exc_type, exc_val, exc_tb):
                return await self.adapter.__aexit__(exc_type, exc_val, exc_tb)

            async def _wrapped_get(self, url, do_request):
                res = None
                gen = self._retries(url)
                while True:
                    try:
                        next(gen)
                    except StopIteration:
                        break
                    assert res is None
                    try:
                        res = await do_request()
                    except Exception as e:
                        error_wait_seconds = gen.throw(e)
                        await asyncio.sleep(error_wait_seconds)
                    else:
                        assert gen.send(res) is None
                assert res is not None
                return res

    else:

        class AdapterProxy(BaseAdapterProxy, BaseSyncAdapter):
            def _wrapped_get(self, url, do_request):
                res = None
                gen = self._retries(url)
                while True:
                    try:
                        next(gen)
                    except StopIteration:
                        break
                    assert res is None
                    try:
                        res = do_request()
                    except Exception as e:
                        error_wait_seconds = gen.throw(e)
                        sleep(error_wait_seconds)
                    else:
                        assert gen.send(res) is None
                assert res is not None
                return res

    # In order to take advantage of Keep-Alives in tests, the actual Adapter
    # should be persisted between the test runs, so this fixture must be
    # in the "session" scope.
    adapter_factory = partial(
        AdapterProxy,
        adapter_factory=default_adapter,
        requests_monitor=requests_monitor,
        is_internet_access_allowed=is_internet_access_allowed,
    )
    with patch.object(
        geopy.geocoders.options, "default_adapter_factory", adapter_factory
    ):
        yield


class BaseAdapterProxy:
    def __init__(
        self,
        *,
        proxies,
        ssl_context,
        adapter_factory,
        requests_monitor,
        is_internet_access_allowed
    ):
        self.adapter = adapter_factory(
            proxies=proxies,
            ssl_context=ssl_context,
        )
        self.requests_monitor = requests_monitor
        self.is_internet_access_allowed = is_internet_access_allowed

    def get_json(self, url, *, timeout, headers):
        return self._wrapped_get(
            url,
            partial(self.adapter.get_json, url, timeout=timeout, headers=headers),
        )

    def get_text(self, url, *, timeout, headers):
        return self._wrapped_get(
            url,
            partial(self.adapter.get_text, url, timeout=timeout, headers=headers),
        )

    def _retries(self, url):
        if not self.is_internet_access_allowed:
            # Assume that *all* geocoders require Internet access
            pytest.skip("Skipping a test requiring Internet access")

        self.requests_monitor.record_request(url)

        netloc = netloc_from_url(url)
        retries = max_retries

        if netloc in no_retries_for_hosts:
            retries = 0

        for i in range(retries + 1):
            try:
                with self.requests_monitor.record_response(url):
                    yield
            except AdapterHTTPError as error:
                if i == retries or error.status_code not in retry_status_codes:
                    # Note: we shouldn't blindly retry on any >=400 code,
                    # because some of them are actually expected in tests
                    # (like input validation verification).

                    # TODO Retry failures with the 200 code?
                    # Some geocoders return failures with 200 code
                    # (like GoogleV3 for Quota Exceeded).
                    # Should we detect this somehow to restart such requests?
                    #
                    # Re-raise -- don't retry this request
                    raise
                else:
                    # Swallow the error and retry the request
                    pass
            except Exception:
                if i == retries:
                    raise
            else:
                yield None
                return

            self.requests_monitor.record_retry(url)
            yield error_wait_seconds
        raise RuntimeError("Should not have been reached")



#!/usr/bin/env python
"""
geopy
"""

import sys

from setuptools import find_packages, setup

if sys.version_info < (3, 5):
    raise RuntimeError(
        "geopy 2 supports Python 3.5 and above. "
        "Use geopy 1.x if you need Python 2.7 or 3.4 support."
    )

# This import must be below the above `sys.version_info` check,
# because the code being imported here is not compatible with the older
# versions of Python.
from geopy import __version__ as version  # noqa  # isort:skip

INSTALL_REQUIRES = [
    'geographiclib<2,>=1.49',
]

EXTRAS_DEV_TESTFILES_COMMON = [
    "async_generator",
]

EXTRAS_DEV_LINT = [
    "flake8>=3.6.0,<3.7.0",
    "isort>=4.3.4,<4.4.0",
]

EXTRAS_DEV_TEST = [
    "coverage",
    "pytest-aiohttp",  # for `async def` tests
    "pytest>=3.10",
]

EXTRAS_DEV_DOCS = [
    "readme_renderer",
    "sphinx",
    "sphinx_rtd_theme>=0.4.0",
]

setup(
    name='geopy',
    version=version,
    description='Python Geocoding Toolbox',
    long_description=open('README.rst').read(),
    maintainer='Kostya Esmukov',
    maintainer_email='kostya@esmukov.ru',
    url='https://github.com/geopy/geopy',
    download_url=(
        'https://github.com/geopy/geopy/archive/%s.tar.gz' % version
    ),
    packages=find_packages(exclude=["*test*"]),
    install_requires=INSTALL_REQUIRES,
    extras_require={
        "dev": (EXTRAS_DEV_TESTFILES_COMMON +
                EXTRAS_DEV_LINT +
                EXTRAS_DEV_TEST +
                EXTRAS_DEV_DOCS),
        "dev-lint": (EXTRAS_DEV_TESTFILES_COMMON +
                     EXTRAS_DEV_LINT),
        "dev-test": (EXTRAS_DEV_TESTFILES_COMMON +
                     EXTRAS_DEV_TEST),
        "dev-docs": EXTRAS_DEV_DOCS,
        "aiohttp": ["aiohttp"],
        "requests": [
            "urllib3>=1.24.2",
            # ^^^ earlier versions would work, but a custom ssl
            # context would silently have system certificates be loaded as
            # trusted: https://github.com/urllib3/urllib3/pull/1566

            "requests>=2.16.2",
            # ^^^ earlier versions would work, but they use an older
            # vendored version of urllib3 (see note above)
        ],
        "timezone": ["pytz"],
    },
    license='MIT',
    keywords='geocode geocoding gis geographical maps earth distance',
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ]
)
